# -*- coding: utf-8 -*-

from typing import Union

import warnings

import numpy as np

import pyqtgraph as pg

from PySide6.QtCore import Qt, Slot, Signal, QModelIndex, QPointF

from PySide6.QtWidgets import (
    QCheckBox, QComboBox, QListView, QAbstractItemView, QSplitter, QToolBar, QLabel,
    QSlider, QVBoxLayout, QHBoxLayout, QWidget, QSpinBox, QSizePolicy
)

from PySide6.QtGui import QIcon, QAction, QGuiApplication

from insarviz.BandSlider import SliderLeftClickStyle

from insarviz.plot.AbstractPlotView import AbstractPlotWindow, AbstractPlotWidget, ProxyNoItemModel

from insarviz.plot.PlotModel import PlotModel

from insarviz.Loader import Loader

from insarviz.map.layers.SelectionLayer import SelectionItem

from insarviz.utils import get_nearest

from insarviz.Roles import Roles

from insarviz.BandSlider import TimelineHandle


# TemporalPlotWidget class #########################################################################

class TemporalPlotWidget(AbstractPlotWidget):
    """
    temporal evolution plot
    """

    def __init__(self, plot_model: PlotModel):
        super().__init__(plot_model)
        # used by insarviz.exporters.myCSVExporter
        self.plotItem.insarviz_widget = self
        # date marker synchronized with application's window slider
        self.date_marker: DateMarker = DateMarker(self.plot_model.loader)
        # pointer curve
        self.pointer_curve: pg.PlotDataItem = pg.PlotDataItem(symbol='o', symbolSize=5,
                                                              antialias=True)
        self.pointer_curve.setPen("r", width=2)
        self.pointer_curve.setSymbolPen(None)
        self.pointer_curve.setSymbolBrush("r")
        self.pointer_curve.opts["name"] = "pointer"
        self.pointer_curve.setZValue(1.)  # interactive curve is plotted in foreground
        # selection items curves
        self.points_curves: list[pg.PlotDataItem] = []
        self.profile_curve: pg.PlotDataItem = pg.PlotDataItem(symbol='o', symbolSize=5,
                                                              antialias=True)
        self.profile_curve.sigPointsHovered.connect(self.on_points_hover)
        self.profile_curve.scatter.setData(hoverable=True, tip=None, hoverBrush='r', hoverSize=8)
        self.profile_curve.setSymbolPen(None)
        self.references_curves: list[pg.PlotDataItem] = []
        self.profile: int = -1  # index of the current profile, -1 means no profile
        self.profile_point: int = 0  # index of the current profile point, 0 means start of the profile
        self.reference: int = -1  # index of the current reference, -1 means no reference
        # signals and slots
        plot_model.loader.data_loaded.connect(self.on_data_loaded)
        plot_model.loader.histograms_computed.connect(self.on_histograms_computed)
        plot_model.closed.connect(self.on_close)
        plot_model.updated_pointer_info.connect(self.plot_pointer_data)
        plot_model.points_added.connect(self.add_points_curves)
        plot_model.points_removed.connect(self.remove_points_curves)
        plot_model.points_updated.connect(self.update_points_curves)
        plot_model.profiles_updated.connect(self.update_profiles_curves)
        plot_model.references_added.connect(self.add_references_curves)
        plot_model.references_removed.connect(self.remove_references_curves)
        plot_model.references_updated.connect(self.update_references_curves)

    @Slot(int)
    def on_data_loaded(self, band_index: int) -> None:
        assert self.plot_model.loader.timestamps is not None
        # x-axis:
        if isinstance(self.plot_model.loader.dates[0], int):
            # dates not available: regular x-axis
            self.plotItem.setAxisItems({'bottom': pg.AxisItem('bottom', text="Band #")})
        else:
            # dates available: x-axis is a TimeAxisItem
            self.plotItem.setAxisItems({'bottom': TimeAxisItem(text='Date', units='yyyy-mm-dd')})
        self.plotItem.setXRange(self.plot_model.loader.timestamps[0],
                                self.plot_model.loader.timestamps[-1], padding=self.default_padding)
        # set the min max range of the x axis
        timedelta = self.plot_model.loader.timestamps[-1] - self.plot_model.loader.timestamps[0]
        xmin = self.plot_model.loader.timestamps[0] - self.max_padding*timedelta
        xmax = self.plot_model.loader.timestamps[-1] + self.max_padding*timedelta
        self.plotItem.setLimits(xMin=xmin, xMax=xmax)
        # date marker
        assert self.plot_model.loader.dataset is not None
        assert band_index in self.plot_model.loader.dataset.indexes
        index = self.plot_model.loader.dataset.indexes.index(band_index)
        self.date_marker.setPos(self.plot_model.loader.timestamps[index])
        self.plotItem.addItem(self.date_marker)
        # pointer curve
        self.plotItem.addItem(self.pointer_curve)

    @Slot()
    def on_histograms_computed(self) -> None:
        assert self.plot_model.loader.total_histogram is not None
        total_bins = self.plot_model.loader.total_histogram[1]
        ydelta = total_bins[-1] - total_bins[0]
        ymin = total_bins[0] - self.max_padding * ydelta
        ymax = total_bins[-1] + self.max_padding * ydelta
        self.plotItem.setLimits(yMin=ymin, yMax=ymax)

    @Slot()
    def on_close(self) -> None:
        self.plotItem.clear()
        del self.points_curves[:]
        #  self.profile_curve is cleared by TemporalPlotWidget.on_close calls
        del self.references_curves[:]
        self.pointer_curve.setData()
        self.pointer_curve.opts["name"] = "pointer"
        self.tooltip_text = {}
        self.plotItem.getViewBox().setToolTip("")

    @Slot(tuple, np.ndarray)
    def plot_pointer_data(self, info: tuple, data: np.ndarray) -> None:
        """
        set pointer curve data from the pixel hovered over on the map
        """
        if info == ():
            self.pointer_curve.setData()
            self.pointer_curve.opts["name"] = "pointer"
        else:
            # test if at least one element of data is not nan
            if not all(np.isnan(data)):
                self.compute_curve(data, self.pointer_curve)
            else:
                self.pointer_curve.setData()
            self.pointer_curve.opts["name"] = f"pointer at ({info[0]}, {info[1]})"

    @Slot(int, int)
    def add_points_curves(self, first: int, last: int) -> None:
        new_curves: list[pg.PlotDataItem] = []
        for i in range(first, last+1):
            point_index: QModelIndex = self.plot_model.points.index(i, 0)
            color = self.plot_model.points.data(point_index, Qt.ItemDataRole.DecorationRole)
            curve: pg.PlotDataItem = pg.PlotDataItem(symbol='o', symbolSize=5, antialias=True)
            curve.sigPointsHovered.connect(self.on_points_hover)
            curve.scatter.setData(hoverable=True, tip=None, hoverBrush='r', hoverSize=8)
            curve.setPen(color, width=2)
            curve.setSymbolPen(None)
            curve.setSymbolBrush(color)
            curve.opts["name"] = self.plot_model.points.data(point_index,
                                                             Qt.ItemDataRole.DisplayRole)
            self.compute_curve(self.plot_model.points_data[i], curve)
            if self.plot_model.points.data(point_index, SelectionItem.ShowCurveRole):
                self.plotItem.addItem(curve)
            new_curves.append(curve)
        self.points_curves[first:first] = new_curves

    @Slot(int, int)
    def remove_points_curves(self, first: int, last: int) -> None:
        for i in range(first, last+1):
            self.plotItem.removeItem(self.points_curves[i])
            # remove tooltip for this curve
            self.points_curves[i].sigPointsHovered.emit(self.points_curves[i], [], None)
        del self.points_curves[first:last+1]

    @Slot(int, int, list)
    def update_points_curves(self, first: int, last: int, roles: list[int]) -> None:
        for i in range(first, last+1):
            point_index = self.plot_model.points.index(i, 0)
            if Roles.ComputeDataRole in roles:
                self.compute_curve(self.plot_model.points_data[i], self.points_curves[i])
                # remove tooltip for this curve
                self.points_curves[i].sigPointsHovered.emit(self.points_curves[i], [], None)
            if Roles.CurveColorRole in roles:
                color = self.plot_model.points.data(point_index, Qt.ItemDataRole.DecorationRole)
                self.points_curves[i].setPen(color, width=2)
                self.points_curves[i].setSymbolPen(None)
                self.points_curves[i].setSymbolBrush(color)
            if Qt.ItemDataRole.EditRole in roles:
                # remove tooltip for this curve
                self.points_curves[i].sigPointsHovered.emit(self.points_curves[i], [], None)
                name = self.plot_model.points.data(point_index, Qt.ItemDataRole.DisplayRole)
                self.points_curves[i].opts["name"] = name
            if SelectionItem.ShowCurveRole in roles:
                show = (self.plot_model.points.data(point_index, SelectionItem.ShowCurveRole) ==
                        Qt.CheckState.Checked)
                if show:
                    with warnings.catch_warnings():
                        # ignore UserWarning for curve already plotted
                        warnings.filterwarnings("ignore", category=UserWarning,
                                                message="Item already added to PlotItem")
                        self.plotItem.addItem(self.points_curves[i])
                else:
                    self.plotItem.removeItem(self.points_curves[i])
                    # remove tooltip for this curve
                    self.points_curves[i].sigPointsHovered.emit(self.points_curves[i], [], None)

    @Slot(int, int, list)
    def update_profiles_curves(self, first: int, last: int, roles: list[int]) -> None:
        if first <= self.profile <= last:
            profile_index = self.plot_model.profiles.index(self.profile, 0)
            if Roles.ComputeDataRole in roles:
                self.compute_curve(
                    self.plot_model.profiles_data[self.profile][self.profile_point, :],
                    self.profile_curve)
                # remove tooltip for this curve
                self.profile_curve.sigPointsHovered.emit(self.profile_curve, [], None)
            if Roles.CurveColorRole in roles:
                color = self.plot_model.profiles.data(profile_index, Qt.ItemDataRole.DecorationRole)
                self.profile_curve.setPen(color, width=2)
                self.profile_curve.setSymbolBrush(color)
            if Qt.ItemDataRole.EditRole in roles:
                # remove tooltip for this curve
                self.profile_curve.sigPointsHovered.emit(self.profile_curve, [], None)
                name = self.plot_model.profiles.data(profile_index, Qt.ItemDataRole.DisplayRole)
                self.profile_curve.opts["name"] = name

    @Slot(int, int)
    def add_references_curves(self, first: int, last: int) -> None:
        new_curves = []
        for i in range(first, last+1):
            reference_index = self.plot_model.references.index(i, 0)
            color = self.plot_model.references.data(reference_index, Qt.ItemDataRole.DecorationRole)
            curve = pg.PlotDataItem(symbol='o', symbolSize=5, antialias=True)
            curve.sigPointsHovered.connect(self.on_points_hover)
            curve.scatter.setData(hoverable=True, tip=None, hoverBrush='r', hoverSize=8)
            curve.setPen(color, width=2)
            curve.setSymbolPen(None)
            curve.setSymbolBrush(color)
            curve.setZValue(-1.)  # reference curves are plotted in background
            curve.opts["name"] = self.plot_model.references.data(reference_index,
                                                                 Qt.ItemDataRole.DisplayRole)
            self.compute_curve(self.plot_model.references_data[i], curve)
            if self.plot_model.references.data(reference_index, SelectionItem.ShowCurveRole):
                self.plotItem.addItem(curve)
            new_curves.append(curve)
        self.references_curves[first:first] = new_curves

    @Slot(int, int)
    def remove_references_curves(self, first: int, last: int) -> None:
        for i in range(first, last+1):
            self.plotItem.removeItem(self.references_curves[i])
            # remove tooltip for this curve
            self.references_curves[i].sigPointsHovered.emit(self.references_curves[i], [], None)
        del self.references_curves[first:last+1]

    @Slot(int, int, list)
    def update_references_curves(self, first: int, last: int, roles: list[int]) -> None:
        for i in range(first, last+1):
            reference_index = self.plot_model.references.index(i, 0)
            if Roles.ComputeDataRole in roles:
                self.compute_curve(self.plot_model.references_data[i], self.references_curves[i])
                # remove tooltip for this curve
                self.references_curves[i].sigPointsHovered.emit(self.references_curves[i], [], None)
            if Roles.CurveColorRole in roles:
                color = self.plot_model.references.data(reference_index,
                                                        Qt.ItemDataRole.DecorationRole)
                self.references_curves[i].setPen(color, width=2)
                self.references_curves[i].setSymbolPen(None)
                self.references_curves[i].setSymbolBrush(color)
            if Qt.ItemDataRole.EditRole in roles:
                # remove tooltip for this curve
                self.references_curves[i].sigPointsHovered.emit(self.references_curves[i], [], None)
                name = self.plot_model.references.data(reference_index, Qt.ItemDataRole.DisplayRole)
                self.references_curves[i].opts["name"] = name
            if SelectionItem.ShowCurveRole in roles:
                show = (self.plot_model.references.data(reference_index, SelectionItem.ShowCurveRole)
                        == Qt.CheckState.Checked)
                if show:
                    with warnings.catch_warnings():
                        # ignore UserWarning for curve already plotted
                        warnings.filterwarnings("ignore", category=UserWarning,
                                                message="Item already added to PlotItem")
                        self.plotItem.addItem(self.references_curves[i])
                else:
                    self.plotItem.removeItem(self.references_curves[i])
                    # remove tooltip for this curve
                    self.references_curves[i].sigPointsHovered.emit(
                        self.references_curves[i], [], None)

    @Slot(int)
    def change_profile(self, i: int) -> None:
        if self.profile != i:
            # remove tooltip for this curve
            self.profile_curve.sigPointsHovered.emit(self.profile_curve, [], None)
            if self.profile != -1:
                profile_index = self.plot_model.profiles.index(self.profile, 0)
                self.plot_model.profiles.setData(profile_index, None,
                                                 Roles.ProfileTemporalRole)
        self.profile = i
        if self.profile != -1:
            profile_index = self.plot_model.profiles.index(self.profile, 0)
            color = self.plot_model.profiles.data(profile_index, Qt.ItemDataRole.DecorationRole)
            self.profile_curve.setPen(color, width=2)
            self.profile_curve.setSymbolBrush(color)
            name = self.plot_model.profiles.data(profile_index, Qt.ItemDataRole.DisplayRole)
            self.profile_curve.opts["name"] = f"{name} at pixel {self.profile_point}"
            self.compute_curve(self.plot_model.profiles_data[self.profile][self.profile_point, :],
                               self.profile_curve)
            self.plotItem.addItem(self.profile_curve)
        else:
            self.profile_curve.opts["name"] = "no profile"
            self.plotItem.removeItem(self.profile_curve)

    @Slot(int)
    def change_profile_point(self, i: int) -> None:
        self.profile_point = i
        if self.profile != -1:
            profile_index = self.plot_model.profiles.index(self.profile, 0)
            self.plot_model.profiles.setData(profile_index, self.profile_point,
                                             Roles.ProfileTemporalRole)
            name = self.plot_model.profiles.data(profile_index, Qt.ItemDataRole.DisplayRole)
            self.profile_curve.opts["name"] = f"{name} at pixel {self.profile_point}"
            self.compute_curve(self.plot_model.profiles_data[self.profile][self.profile_point, :],
                               self.profile_curve)

    @Slot(int)
    def change_reference(self, i: int) -> None:
        old_reference: Union[int, np.ndarray]
        if self.reference != -1:
            old_reference = self.plot_model.references_data[self.reference]
        else:
            old_reference = 0
        self.reference = i
        # pointer curve
        if self.pointer_curve.getData() != (None, None):
            self.compute_curve(self.pointer_curve.getData()[1] - old_reference, self.pointer_curve)
        # points
        for k, point_curve in enumerate(self.points_curves):
            self.compute_curve(self.plot_model.points_data[k], point_curve)
        # references
        for k, ref_curve in enumerate(self.references_curves):
            self.compute_curve(self.plot_model.references_data[k], ref_curve)
        # profiles
        if self.profile != -1:
            self.compute_curve(self.plot_model.profiles_data[self.profile][self.profile_point, :],
                               self.profile_curve)

    def compute_curve(self, data: np.ndarray, curve: pg.PlotDataItem,
                      use_reference: bool = True) -> None:
        x = self.plot_model.loader.timestamps
        if self.reference == -1 or not use_reference:
            y = data
        else:
            y = data - self.plot_model.references_data[self.reference]
        curve.setData(x, y)


# TemporalPlotWindow class #########################################################################

class TemporalPlotWindow(AbstractPlotWindow):

    reference_changed = Signal(int)
    profile_changed = Signal(int)

    def __init__(self, plot_model: PlotModel):
        super().__init__(plot_model, TemporalPlotWidget(plot_model))
        plot_model.closed.connect(self.on_close)
        self.setWindowTitle("Temporal profile")
        # checkbox to fix axes when hovering on Map:
        self.toolbar.addSeparator()
        self.checkbox_axes: QCheckBox = QCheckBox('Lock axes', self)
        self.checkbox_axes.setToolTip("When checked, keep axes ranges while hovering on Map."
                                      " Note that you can still zoom in/out on the plot.")
        self.checkbox_axes.stateChanged.connect(self.ctrl_axes)
        self.toolbar.addWidget(self.checkbox_axes)
        # reference icon
        self.toolbar.addSeparator()
        reference_icon: QLabel = QLabel()
        reference_icon.setPixmap(QIcon('icons:ref.png').pixmap(self.toolbar.iconSize()))
        self.toolbar.addWidget(reference_icon)
        # combobox to choose reference
        self.reference_combobox: QComboBox = QComboBox(self)
        proxy_reference_model = ProxyNoItemModel(self.plot_model.references, "No reference")
        self.reference_combobox.setModel(proxy_reference_model)
        proxy_reference_model.rowsAboutToBeRemoved.connect(
            self.update_reference_combobox_index_before_remove)
        self.reference_combobox.currentIndexChanged.connect(self.reference_combobox_index_changed)
        self.reference_changed.connect(self.plot_widget.change_reference)
        self.toolbar.addWidget(self.reference_combobox)
        # spacer
        self.spacer = QWidget()
        self.spacer.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Preferred)
        self.toolbar.addWidget(self.spacer)
        # export button
        self.toolbar.addWidget(self.export_button)
        # points show all action
        self.points_showall_action: QAction = QAction(QIcon('icons:eye_open.svg'), "Show all", self)
        self.points_showall_action.setToolTip("Show all point curves")
        self.points_showall_action.triggered.connect(self.plot_model.points.show_all_curves)
        # points hide all action
        self.points_hideall_action: QAction = QAction(QIcon('icons:eye_closed.svg'), "Hide all",
                                                      self)
        self.points_hideall_action.setToolTip("Hide all point curves")
        self.points_hideall_action.triggered.connect(self.plot_model.points.hide_all_curves)
        # points toolbar
        points_toolbar: QToolBar = QToolBar(self)
        points_icon: QLabel = QLabel()
        points_icon.setPixmap(QIcon('icons:points.png').pixmap(points_toolbar.iconSize()))
        points_toolbar.addWidget(points_icon)
        points_toolbar.addWidget(QLabel(" Points :"))
        points_toolbar.addAction(self.points_showall_action)
        points_toolbar.addAction(self.points_hideall_action)
        # points view
        self.points_view: QListView = QListView(parent=self)
        self.points_view.setSelectionMode(QAbstractItemView.SelectionMode.NoSelection)
        self.points_view.setModel(self.plot_model.points)
        # points widget
        points_widget: QWidget = QWidget()
        points_layout: QVBoxLayout = QVBoxLayout()
        points_layout.addWidget(points_toolbar)
        points_layout.addWidget(self.points_view)
        points_widget.setLayout(points_layout)
        # combobox to choose profile
        self.profile_combobox: QComboBox = QComboBox(self)
        self.profile_combobox.setSizePolicy(QSizePolicy.Policy.MinimumExpanding,
                                            self.profile_combobox.sizePolicy().verticalPolicy())
        proxy_profile_model = ProxyNoItemModel(self.plot_model.profiles, "No profile")
        self.profile_combobox.setModel(proxy_profile_model)
        proxy_profile_model.rowsAboutToBeRemoved.connect(
            self.update_profile_combobox_index_before_remove)
        self.profile_combobox.currentIndexChanged.connect(self.profile_combobox_index_changed)
        self.profile_changed.connect(self.plot_widget.change_profile)
        # profiles toolbar
        profiles_toolbar: QToolBar = QToolBar(self)
        profiles_icon: QLabel = QLabel()
        profiles_icon.setPixmap(QIcon('icons:profile.png').pixmap(profiles_toolbar.iconSize()))
        profiles_toolbar.addWidget(profiles_icon)
        profiles_toolbar.addWidget(self.profile_combobox)
        # spinbox to choose profile point
        self.profile_spinbox = QSpinBox(self)
        self.profile_spinbox.setKeyboardTracking(False)
        self.profile_spinbox.setDisabled(True)
        # slider to choose profile point
        self.profile_slider: QSlider = QSlider(Qt.Orientation.Horizontal, self)
        self.profile_slider.setStyle(SliderLeftClickStyle())
        self.profile_slider.setDisabled(True)
        self.profile_spinbox.valueChanged.connect(self.profile_slider.setValue)
        self.profile_slider.valueChanged.connect(self.profile_spinbox.setValue)
        self.profile_slider.valueChanged.connect(self.plot_widget.change_profile_point)
        # profile label
        self.profile_label: QLabel = QLabel("position along profile in pixels")
        self.profile_label.setAlignment(Qt.AlignmentFlag.AlignHCenter)
        self.profile_label.setDisabled(True)
        # profiles widget
        profiles_widget: QWidget = QWidget()
        profiles_layout: QVBoxLayout = QVBoxLayout()
        profiles_layout.addWidget(profiles_toolbar)
        profiles_slider_layout: QHBoxLayout = QHBoxLayout()
        profiles_slider_layout.addWidget(self.profile_spinbox)
        profiles_slider_layout.addWidget(self.profile_slider)
        profiles_layout.addLayout(profiles_slider_layout)
        profiles_layout.addWidget(self.profile_label)
        profiles_widget.setLayout(profiles_layout)
        profiles_widget.setMaximumHeight(profiles_widget.sizeHint().height())
        # references show all action
        self.references_showall_action: QAction = QAction(QIcon('icons:eye_open.svg'), "Show all",
                                                          self)
        self.references_showall_action.setToolTip("Show all reference curves")
        self.references_showall_action.triggered.connect(self.plot_model.references.show_all_curves)
        # references hide all action
        self.references_hideall_action: QAction = QAction(QIcon('icons:eye_closed.svg'), "Hide all",
                                                          self)
        self.references_hideall_action.setToolTip("Hide all reference curves")
        self.references_hideall_action.triggered.connect(self.plot_model.references.hide_all_curves)
        # references toolbar
        references_toolbar: QToolBar = QToolBar(self)
        references_icon: QLabel = QLabel()
        references_icon.setPixmap(QIcon('icons:ref.png').pixmap(points_toolbar.iconSize()))
        references_toolbar.addWidget(references_icon)
        references_toolbar.addWidget(QLabel(" References :"))
        references_toolbar.addAction(self.references_showall_action)
        references_toolbar.addAction(self.references_hideall_action)
        # references view
        self.references_view: QListView = QListView(parent=self)
        self.references_view.setSelectionMode(QAbstractItemView.SelectionMode.NoSelection)
        self.references_view.setModel(self.plot_model.references)
        # references widget
        references_widget: QWidget = QWidget()
        references_layout: QVBoxLayout = QVBoxLayout()
        references_layout.addWidget(references_toolbar)
        references_layout.addWidget(self.references_view)
        references_widget.setLayout(references_layout)
        # layout
        h_splitter: QSplitter = QSplitter(Qt.Orientation.Horizontal)
        h_splitter.setChildrenCollapsible(False)
        h_splitter.addWidget(self.plot_widget)
        v_splitter: QSplitter = QSplitter(Qt.Orientation.Vertical)
        v_splitter.setChildrenCollapsible(False)
        h_splitter.addWidget(v_splitter)
        v_splitter.addWidget(points_widget)
        v_splitter.addWidget(profiles_widget)
        v_splitter.addWidget(references_widget)
        self.layout().addWidget(h_splitter)

    @Slot(int)
    def ctrl_axes(self, state: int) -> None:
        if state == Qt.CheckState.Checked.value:
            self.plot_widget.plotItem.autoBtn.mode = 'fix'
        else:
            self.plot_widget.plotItem.autoBtn.mode = 'auto'
        self.plot_widget.plotItem.autoBtnClicked()

    @Slot()
    def on_close(self) -> None:
        self.profile_combobox.setCurrentIndex(-1)
        self.profile_slider.setValue(0)
        self.reference_combobox.setCurrentIndex(-1)
        self.checkbox_axes.setChecked(False)

    @Slot(int)
    def reference_combobox_index_changed(self, i: int) -> None:
        if i == -1:
            # combobox is empty so we reset it with the "no reference" item
            self.reference_combobox.setCurrentIndex(0)
        else:
            # for TemporalPlotWidget.reference, -1 means no reference, 0 means first reference...
            self.reference_changed.emit(i-1)

    @Slot(int)
    def profile_combobox_index_changed(self, i: int) -> None:
        if i == -1:
            # combobox is empty so we reset it with the "no reference" item
            self.profile_combobox.setCurrentIndex(0)
        else:
            if i == 0:
                self.profile_slider.setTickPosition(QSlider.TickPosition.NoTicks)
                self.profile_slider.setMinimum(0)
                self.profile_slider.setMaximum(0)
                self.profile_slider.setDisabled(True)
                self.profile_spinbox.setDisabled(True)
                self.profile_label.setDisabled(True)
            else:
                length: int = len(self.plot_model.profiles_data[i-1])
                if length <= 40:
                    self.profile_slider.setTickPosition(QSlider.TickPosition.TicksBelow)
                else:
                    self.profile_slider.setTickPosition(QSlider.TickPosition.NoTicks)
                self.profile_slider.setTickInterval(1)
                self.profile_slider.setMinimum(0)
                self.profile_spinbox.setMinimum(0)
                self.profile_slider.setMaximum(length-1)
                self.profile_spinbox.setMaximum(length-1)
                self.profile_slider.setEnabled(True)
                self.profile_spinbox.setEnabled(True)
                self.profile_label.setEnabled(True)
            # for TemporalPlotWidget.profile, -1 means no profile, 0 means first profile...
            self.profile_changed.emit(i-1)
            self.profile_slider.setValue(0)
            self.profile_slider.valueChanged.emit(0)

    @Slot(QModelIndex, int, int)
    def update_reference_combobox_index_before_remove(self, parent: QModelIndex, first: int,
                                                      last: int) -> None:
        # pylint: disable=unused-argument
        previous_index = self.reference_combobox.currentIndex()
        if previous_index >= first and previous_index <= last:
            # otherwise QComboBox set currentIndex to first-1
            self.reference_combobox.setCurrentIndex(0)

    @Slot(QModelIndex, int, int)
    def update_profile_combobox_index_before_remove(self, parent: QModelIndex, first, last) -> None:
        # pylint: disable=unused-argument
        previous_index = self.profile_combobox.currentIndex()
        if previous_index >= first and previous_index <= last:
            # otherwise QComboBox set currentIndex to first-1
            self.profile_combobox.setCurrentIndex(0)


# TimeAxisItem class ###############################################################################

class TimeAxisItem(pg.DateAxisItem):
    """
    Based on pyqtgraph's DateAxisItem.
    Axis item that displays dates from unix timestamps, SI prefix
    for units is disabled and values are handled to be displayed in
    yyyy-mm-dd format.
    """

    def __init__(self, text: str, units: str):
        """
        Creates a new TimeAxisItem.

        Parameters
        ----------
        text : str
            text (without units) to display on the label for this axis
        units : str
            units for this axis
        """
        super().__init__()
        # to avoid unit scaling (unfit for dates) :
        self.enableAutoSIPrefix(False)
        self.setLabel(text=text, units=units)


# DateMarker class #################################################################################

class DateMarker(pg.InfiniteLine):
    """
    Based on pyqtgraph's InfiniteLine
    Vertical infinite line on plot, x-axis-position is synchronized with
    MainWindow's slider position
    overload some methods
    """

    pos_changed = Signal(int)

    base_pen = pg.functions.mkPen(0.5, width=3)
    dragged_pen = TimelineHandle.dragged_pen
    hover_pen = TimelineHandle.hover_pen

    def __init__(self, loader: Loader):
        self.loader: Loader = loader
        super().__init__(pen=DateMarker.base_pen, hoverPen=DateMarker.hover_pen, angle=90,
                         movable=True)
        self.setZValue(-100)
        self.setCursor(Qt.CursorShape.OpenHandCursor)
        self.sigDragged.connect(lambda _: self.setPen(DateMarker.dragged_pen))
        self.sigPositionChangeFinished.connect(lambda _: self.setPen(DateMarker.base_pen))

    def setPos(self, pos: Union[float, QPointF]) -> None:
        """ 
        Overload function
        Position can only be set to one of the values of the dataset
        (on x-axis, ie a band# or date)

        Parameters
        ----------
        pos : int, float, QPoint
            position (approx.) where the line should be set
        """
        if isinstance(pos, QPointF):
            pos = pos.x()
        if self.loader.timestamps is not None and pos not in self.loader.timestamps:
            pos, _ = get_nearest(self.loader.timestamps, pos)
        super().setPos(pos)

    def mouseDragEvent(self, ev):
        """
        Overload function
        emit signal that position changed, to update the position of the slider
        in MainWindow accordingly

        Parameters
        ----------
        ev : QMouseEvent
        """
        super().mouseDragEvent(ev)
        if ev.isAccepted():
            _, idx = get_nearest(self.loader.timestamps, self.getPos()[0])
            self.pos_changed.emit(self.loader.dataset.indexes[idx])
            if ev.isStart():
                QGuiApplication.instance().setOverrideCursor(Qt.CursorShape.DragMoveCursor)
            elif ev.isFinish():
                QGuiApplication.instance().restoreOverrideCursor()

    @Slot(int)
    def on_slider_changed(self, slidervalue: int) -> None:
        """
        Receive signal when MainWindow's slider position changed, update
        line position accordingly

        Parameters
        ----------
        slidervalue : int
            New position of slider tick.
        """
        assert self.loader.dataset is not None and self.loader.timestamps is not None
        self.setPos(self.loader.timestamps[self.loader.dataset.indexes.index(slidervalue)])



# -*- coding: utf-8 -*-

from typing import Optional

import logging

from OpenGL import GL

from PySide6.QtCore import Qt, QSize

from PySide6.QtWidgets import (
    QDialog, QWidget, QDialogButtonBox, QVBoxLayout, QSplitter, QSizePolicy
)

from PySide6.QtOpenGLWidgets import QOpenGLWidget

from PySide6.QtGui import QMatrix4x4

from PySide6.QtOpenGL import QOpenGLVertexArrayObject, QOpenGLShaderProgram, QOpenGLTexture

from shiboken6 import VoidPtr

from pyqtgraph import BarGraphItem

from pyqtgraph.colormap import ColorMap

import numpy as np

from insarviz.colormaps import create_colormap_texture

from insarviz.map.Shaders import DATA_UNIT, COLORMAP_UNIT

from insarviz.map.layers.Layer import Raster1BLayer

from insarviz.ColormapWidget import ColormapWidget, myHistogramWidget

from insarviz.linalg import matrix

from insarviz.map.AbstractMapView import AbstractMapView

logger = logging.getLogger(__name__)


class RasterLayerColorMapEditor(QDialog):

    default_padding = ColormapWidget.default_padding
    max_padding = ColormapWidget.max_padding
    autorange_threshold = ColormapWidget.autorange_threshold

    def __init__(self, parent: Optional[QWidget] = None):
        super().__init__(parent)
        self.setWindowModality(Qt.WindowModality.ApplicationModal)
        self.setModal(True)
        self.setWindowTitle("Edit colormap")
        self.setSizeGripEnabled(True)
        self.setSizePolicy(QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding))

        self.colormap: Optional[ColorMap] = None

        self.histogram_widget = myHistogramWidget()
        # store the colormap itself because its name is not saved in gradient
        self.histogram_widget.gradient.menu.sigColorMapTriggered.connect(
            self.store_colormap)
        # new histogram curve
        self.hist_plot = BarGraphItem(x0=[], x1=[], y0=[], y1=[],
                                      pen=None, brush=(220, 20, 20, 100))
        self.hist_plot.setZValue(10)
        self.hist_plot.setRotation(90)
        self.histogram_widget.vb.addItem(self.hist_plot)

        self.layer_view = RasterLayerView()
        self.histogram_widget.sigLevelsChanged.connect(self.on_levels_changed)
        self.histogram_widget.gradient.menu.sigColorMapTriggered.connect(
            self.layer_view.set_colormap)

        self.button_box = QDialogButtonBox()
        self.autorange_button = self.button_box.addButton("Autorange",
                                                          QDialogButtonBox.ButtonRole.ResetRole)
        self.autorange_button.clicked.connect(self.autorange)
        self.cancel_button = self.button_box.addButton(QDialogButtonBox.StandardButton.Cancel)
        self.cancel_button.clicked.connect(self.reject)
        self.apply_button = self.button_box.addButton(QDialogButtonBox.StandardButton.Apply)
        self.apply_button.clicked.connect(self.accept)

        self.main_widget = QSplitter(Qt.Orientation.Horizontal)
        self.main_widget.setChildrenCollapsible(False)
        self.main_widget.addWidget(self.layer_view)
        self.main_widget.addWidget(self.histogram_widget)

        self.main_layout = QVBoxLayout()
        self.main_layout.addWidget(self.main_widget)
        self.main_layout.addWidget(self.button_box)
        self.setLayout(self.main_layout)

    def set_layer(self, layer: Raster1BLayer):
        assert isinstance(layer, Raster1BLayer)
        self.layer_view.set_texture(layer.textures[GL.GL_TEXTURE0+DATA_UNIT], layer.model_matrix)
        self.histogram_widget.setLevels(layer.colormap_v0, layer.colormap_v1)
        self.set_colormap(layer.colormap.name)
        hist, bins = layer.histogram
        self.hist_plot.setOpts(x0=bins[:-1], x1=bins[1:], y0=np.zeros(len(hist)), y1=hist)
        ymin = float(bins[0] - (bins[-1] - bins[0]) * self.max_padding)
        ymax = float(bins[-1] + (bins[-1] - bins[0]) * self.max_padding)
        self.histogram_widget.vb.setLimits(yMin=ymin, yMax=ymax)
        self.histogram_widget.vb.setYRange(bins[1], bins[-2], padding=self.default_padding)

    def on_levels_changed(self) -> None:
        self.layer_view.set_v0_v1(*self.histogram_widget.getLevels())

    def get_v0_v1(self) -> tuple[float, float]:
        return self.histogram_widget.getLevels()

    def autorange(self) -> None:
        hist = self.hist_plot.opts.get('y1')
        bins = np.empty(len(hist)+1)
        bins[:-1] = self.hist_plot.opts.get('x0')
        bins[-1] = self.hist_plot.opts.get('x1')[-1]
        v0, v1 = self.autorange_from_hist(hist, bins)
        self.histogram_widget.setLevels(v0, v1)

    autorange_from_hist = ColormapWidget.autorange_from_hist

    set_colormap = ColormapWidget.set_colormap

    def store_colormap(self, cmap: ColorMap) -> None:
        self.colormap = cmap

    def get_colormap(self) -> ColorMap:
        return self.colormap

    def set_default_colormap(self) -> None:
        # greyscale is the first colormap action (see colormaps.py)
        self.histogram_widget.gradient.menu.actions()[0].trigger()


class RasterLayerView(QOpenGLWidget):

    def __init__(self, parent: QWidget = None):
        super().__init__(parent)
        self.setMinimumSize(QSize(640, 480))
        self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        self.texture = QOpenGLTexture(QOpenGLTexture.Target.Target2D)
        self.colormap_v0: float = 0.
        self.colormap_v1: float = 1.
        self.colormap_texture = QOpenGLTexture(QOpenGLTexture.Target.Target1D)
        self.program: QOpenGLShaderProgram
        # OpenGL vertex array object holding a square on which to display texture
        self.vao = QOpenGLVertexArrayObject()
        self.z: float = 1.0  # zoom level
        # center of the view matrix in data coordinates
        self.cx: float = 0.
        self.cy: float = 0.
        self.model_matrix: matrix.Matrix = matrix.identity()
        self.view_matrix: matrix.Matrix = matrix.identity()
        self.projection_matrix: matrix.Matrix = matrix.identity()

    def initializeGL(self) -> None:
        AbstractMapView.initializeGL(self)
        self.program = QOpenGLShaderProgram()
        Raster1BLayer.build_program(self)

    resizeGL = AbstractMapView.resizeGL

    def paintGL(self) -> None:
        """
        Generate and display OpenGL texture for Map.
        """
        glfunc = self.context().functions()
        glfunc.glClear(GL.GL_COLOR_BUFFER_BIT)
        self.vao.bind()
        # bind textures to texture units
        glfunc.glActiveTexture(GL.GL_TEXTURE0+DATA_UNIT)
        self.texture.bind()
        glfunc.glActiveTexture(GL.GL_TEXTURE0+COLORMAP_UNIT)
        self.colormap_texture.bind()
        self.program.bind()
        # set view and projection matrixes
        self.program.setUniformValue(self.program.uniformLocation('model_matrix'),
                                     QMatrix4x4(matrix.flatten(self.model_matrix)))
        self.program.setUniformValue(self.program.uniformLocation('view_matrix'),
                                     QMatrix4x4(matrix.flatten(self.view_matrix)))
        self.program.setUniformValue(self.program.uniformLocation('projection_matrix'),
                                     QMatrix4x4(matrix.flatten(self.projection_matrix)))
        self.program.setUniformValue1f(self.program.uniformLocation('v0'), float(self.colormap_v0))
        self.program.setUniformValue1f(self.program.uniformLocation('v1'), float(self.colormap_v1))
        # draw the two triangles of the VAO that form a square
        glfunc.glDrawElements(GL.GL_TRIANGLES, 6, GL.GL_UNSIGNED_INT, VoidPtr(0))
        self.vao.release()
        self.texture.release()
        self.colormap_texture.release()
        self.program.release()

    def update_view_matrix(self) -> None:
        # texture size
        tex_w, tex_h = self.model_matrix[0][0], self.model_matrix[1][1]
        self.cx, self.cy = tex_w//2, tex_h//2
        # view size
        view_w, view_h = self.width(), self.height()
        # size ratios
        view_ratio, tex_ratio = view_w / view_h, tex_w / tex_h
        if view_ratio > tex_ratio:
            # view is wider than texture
            # zoom is thus the height ratio between view and texture
            self.z = view_h / tex_h
        else:
            # view is higher than texture
            # zoom is thus the width ratio between view and texture
            self.z = view_w / tex_w
        self.view_matrix = matrix.product(
            matrix.translate(self.width()//2, self.height()//2),
            matrix.scale(self.z, self.z),
            matrix.translate(-self.cx, -self.cy)
        )

    update_projection_matrix = AbstractMapView.update_projection_matrix

    def set_texture(self, texture: QOpenGLTexture, model_matrix: matrix.Matrix) -> None:
        self.texture = texture
        self.model_matrix = matrix.scale(model_matrix[0][0], model_matrix[1][1])

    def set_v0_v1(self, v0: float, v1: float) -> None:
        self.colormap_v0 = float(v0)
        self.colormap_v1 = float(v1)
        self.repaint()

    def set_colormap(self, colormap: ColorMap) -> None:
        self.makeCurrent()
        self.colormap_texture.destroy()
        self.colormap_texture = create_colormap_texture(colormap)
        self.doneCurrent()
        self.repaint()



# -*- coding: utf-8 -*-

"""
LayerModel

This module contains the LayerModel class managing the layers of the Map view.
"""

# imports ##########################################################################################

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional, Iterable

import gc

from PySide6.QtCore import (
    Qt, Slot, Signal, QModelIndex, QMimeData, QByteArray
)

from PySide6.QtGui import QUndoCommand

from insarviz.map.TreeModel import TreeModel, TreeItem

from insarviz.Roles import Roles

from insarviz.map.layers.Layer import Layer

if TYPE_CHECKING:
    from insarviz.map.SelectionLayer import (
        SelectionLayer, SelectionPoint, SelectionProfile, SelectionReference)


# LayerModel class #################################################################################

class LayerModel(TreeModel):
    """
    Model class managing the layers. Follows a tree structure, with first column being the items
    themselves and second column being remove buttons. The children of the root are the Layers.
    Children of layers can be TreeItemAttribute that provide user-friendly access to a TreeItem's
    attribute, or SelectionFolder and recursively SelectionItem in the case of the SelectionLayer.
    Only layer items (at the root of the tree) can be dragged and dropped. An item's checkbox
    manages the visibility of the item in MapView (displayed if checked, hidden if unchecked).
    All actions modifying LayerModel (add, remove, move or modify item) are wrapped inside a
    QUndoCommand that is pushed on the ts_viz.MainWindow's QUndoStack.
    """

    # connected to ts_viz.MainWindow.layer_remove_action.setEnabled
    current_removable = Signal(bool)
    # connected to ts_viz.MainWindow.layer_moveup_action.setEnabled
    current_movable_up = Signal(bool)
    # connected to ts_viz.MainWindow.layer_movedown_action.setEnabled
    current_movable_down = Signal(bool)
    # connected to PlotModel.on_selection_init and to ts_viz.MainWindow.undo_stack.clear
    selection_initialized = Signal(object)  # object = self
    # connected to LayerView.expand
    request_expand = Signal(QModelIndex)
    # connected to ts_viz.MainWindow.undo_stack.push
    add_undo_command = Signal(QUndoCommand)
    # connected to MapModel.request_paint
    request_paint = Signal()

    def __init__(self, layers: Optional[list[Layer]] = None):
        super().__init__()
        if layers is not None:
            for layer in layers:
                self.add_layer(layer)
        # direct access to layers
        self.layers: list[Layer] = self.root_item.child_items
        self.selection: Optional[SelectionLayer] = None

    def columnCount(self, parent: QModelIndex = QModelIndex()) -> int:
        # first column is the items themselves (layer, selection item, etc.)
        # second column is the remove buttons
        return 2

    def setData(self, index: QModelIndex, value: Any, role: int = Qt.ItemDataRole.EditRole) -> bool:
        result: bool = super().setData(index, value, role)
        editor_role: str = index.data(Roles.EditorRole)
        if result is True and editor_role is not None:
            # if editor_role is not None then index is a TreeItemAttribute, and we need to signal
            # that index.parent() has been modified
            roles: list[int] = [Qt.ItemDataRole.EditRole]
            # get the data role of the TreeItemAttribute, that is an indication of what has changed
            # in its parent (color, name, position...)
            data_role: Optional[int] = index.data(Roles.DataRole)
            if data_role is not None:
                roles += [data_role]
            # emit a dataChanged signal of what has changed in the parent (catched by PlotModel
            # for example)
            self.dataChanged.emit(index.parent(), index.parent(), roles)
        return result

    def clear(self) -> None:
        super().clear()
        # force garbage collection
        gc.collect()

    # reimplemented methods for drag and drop
    # multiple selection is disabled, so indexes is assumed to contain an unique index

    def mimeTypes(self) -> list[str]:
        # tiff and insarviz.layer drops are accepted
        return ['image/tiff', 'application/insarviz.layer']

    def mimeData(self, indexes: Iterable[QModelIndex]) -> QMimeData:
        indexes = iter(indexes)
        try:
            index = next(indexes)
        except StopIteration:
            raise RuntimeError("LayerModel: empty drag")  # pylint: disable=raise-missing-from
        try:
            next(indexes)
            raise RuntimeError("LayerModel: drag of more than one row is forbidden")
        except StopIteration:
            pass
        assert index.isValid() and not index.parent().isValid(), \
            "LayerModel : drag of a row that is not a top child (i.e. layer) is forbidden"
        mime_data: QMimeData = QMimeData()
        # store the row number in mimeData.text
        mime_data.setText(str(index.row()))
        # create an 'application/insarviz.layer' entry in mimeData.formats using a dummy QByteArray
        mime_data.setData('application/insarviz.layer', QByteArray())
        return mime_data

    def supportedDropActions(self) -> Qt.DropActions:
        return Qt.DropAction.CopyAction | Qt.DropAction.MoveAction

    def canDropMimeData(self, data: QMimeData, action: Qt.DropAction, row: int, column: int,
                        parent: QModelIndex) -> bool:
        # intersection between data.formats and self.mimeTypes() is empty
        # or intersection between action and self.supportedDropActions() is empty
        if (not set(data.formats()) & set(self.mimeTypes()) or
                not action & self.supportedDropActions()):
            return False
        # dropped on root or directly on layer item
        if not parent.isValid() or (not parent.parent().isValid() and row == -1 and column == -1):
            if data.hasFormat('application/insarviz.layer') and action == Qt.DropAction.MoveAction:
                return True
            elif data.hasFormat('image/tiff') and action == Qt.DropAction.CopyAction:
                return True
            else:
                return False
        else:
            return False

    def dropMimeData(self, data: QMimeData, action: Qt.DropAction, row: int, column: int,
                     parent: QModelIndex) -> bool:
        if not self.canDropMimeData(data, action, row, column, parent):
            return False
        if action == Qt.DropAction.IgnoreAction:
            return True
        if column > 0:
            column = 0
        if parent.isValid():
            # dropping on an item (i.e. not on root)
            if not parent.parent().isValid() and row == -1:
                # dropping on a layer
                row = parent.row() + 1
                parent = QModelIndex()  # i.e. root
            else:
                return False
        elif row == -1:
            # dropping on root and not between layers
            row = self.rowCount(parent)
        if data.hasFormat('application/insarviz.layer'):
            # "data.layer" is the row that we are dragging and "row" is the row we are dropping on
            self.add_undo_command.emit(MoveLayerCommand(self, int(data.text()), row))
            return True
        return False

    # custom methods

    @Slot(object, int)
    def add_layer(self, layer: Layer, i: int = -1) -> None:
        """
        Add layer at i-th row. If i=-1 append layer at the end.
        """
        self.add_undo_command.emit(AddLayerCommand(self, layer, i))

    # connected to LayerView.selectionModel().currentChanged
    @Slot(QModelIndex)
    def manage_current_changed(self, index: QModelIndex) -> None:
        """
        Reacts to change of current selected item in LayerView by emiting signals about whether
        the new selected item is removable / movable up / movable down.
        """
        index = index.siblingAtColumn(self.remove_column)
        if not index.isValid():
            # root item
            self.current_removable.emit(False)
            self.current_movable_up.emit(False)
            self.current_movable_down.emit(False)
        else:
            if index.flags() & Qt.ItemFlag.ItemIsEditable:
                self.current_removable.emit(True)
            else:
                self.current_removable.emit(False)
            if not index.parent().isValid():
                # layer item (parent is root)
                if index.row() == 0:
                    self.current_movable_up.emit(False)
                else:
                    self.current_movable_up.emit(True)
                if index.row() == self.rowCount(index.parent()) - 1:
                    self.current_movable_down.emit(False)
                else:
                    self.current_movable_down.emit(True)
            else:
                self.current_movable_up.emit(False)
                self.current_movable_down.emit(False)

    # connected to MainWindow.layer_showall_action.triggered
    @Slot()
    def show_all_layers(self) -> None:
        for i in range(self.root_item.child_count()):
            index: QModelIndex = self.index(i, 0, QModelIndex())
            self.setData(index, Qt.CheckState.Checked, Qt.ItemDataRole.CheckStateRole)

    # connected to MainWindow.layer_hideall_action.triggered
    @Slot()
    def hide_all_layers(self) -> None:
        for i in range(self.root_item.child_count()):
            index: QModelIndex = self.index(i, 0, QModelIndex())
            self.setData(index, Qt.CheckState.Unchecked, Qt.ItemDataRole.CheckStateRole)

    # connected to MainWindow.layer_moveup_action.triggered
    @Slot(QModelIndex)
    def move_layer_up(self, index: QModelIndex) -> None:
        """
        Move up the layer at index (towards row 0 which is the uppest row).
        """
        if index.parent() == QModelIndex() and index.row() != 0:
            self.add_undo_command.emit(MoveLayerCommand(self, index.row(), index.row()-1))

    # connected to MainWindow.layer_movedown_action.triggered
    @Slot(QModelIndex)
    def move_layer_down(self, index: QModelIndex) -> None:
        """
        Move down the layer at index (towards row self.rowCount(index.parent()) which is the
        lowest row).
        """
        if index.parent() == QModelIndex() and index.row() != self.rowCount(index.parent()) - 1:
            self.add_undo_command.emit(MoveLayerCommand(self, index.row(), index.row()+2))

    # connected to MainWindow.layer_remove_action.triggered, also called by LayerView.remove
    @Slot(QModelIndex)
    def remove(self, index: QModelIndex) -> None:
        """
        Remove the item at index.
        """
        self.add_undo_command.emit(RemoveTreeItemCommand(self, index.row(), index.parent()))

    # called by MapModel after adding selection layer
    def set_selection(self, selection: SelectionLayer) -> None:
        self.selection = selection
        self.selection_initialized.emit(self)

    def add_selection_point(self, point: SelectionPoint) -> None:
        """
        Add a selection point at the end of the point folder.
        """
        self.add_undo_command.emit(AddSelectionPointCommand(self, point))

    def add_selection_profile(self, profile: SelectionProfile) -> None:
        """
        Add a selection profile at the end of the profile folder.
        """
        self.add_undo_command.emit(AddSelectionProfileCommand(self, profile))

    def add_selection_reference(self, reference: SelectionReference) -> None:
        """
        Add a selection reference at the end of the reference folder.
        """
        self.add_undo_command.emit(AddSelectionReferenceCommand(self, reference))

    def get_lineage(self, index: QModelIndex) -> list[int]:
        """
        Return the lineage of index: the child numbers of the path to reach it from the root.
        For example, if index is the third child of the first child of the root:
        self.get_lineage(index) = [0, 2]

        Used to store the index of an item in QUndoCommands. Indeed QModelIndex shall not be stored
        and QPersistentIndex is no longer valid if the item has been removed (even if this remove
        is later undone).
        """
        lineage: list[int] = []
        while index.isValid():
            lineage.insert(0, index.row())
            index = index.parent()
        return lineage

    def get_index_from_lineage(self, lineage: Iterable[int]) -> QModelIndex:
        """
        Return a QModelIndex from a lineage created by LayerModel.get_lineage.
        """
        index: QModelIndex = QModelIndex()
        for i in lineage:
            index = self.index(i, 0, index)
        return index


# QUndoCommand classes #############################################################################

class SetDataCommand(QUndoCommand):

    def __init__(self, layer_model: LayerModel, index: QModelIndex, value: Any, role: int,
                 parent_command: Optional[QUndoCommand] = None,
                 old_value: Any = None):
        super().__init__(parent_command)
        if index.data(Roles.EditorRole) is not None:
            # index is a TreeItemAttribute
            self.setText(
                f"Change {index.data(Qt.ItemDataRole.ToolTipRole)} of {index.parent().data(Qt.ItemDataRole.DisplayRole)}")
        else:
            self.setText(f"Change name of {index.data(Qt.ItemDataRole.DisplayRole)}")
        self.layer_model: LayerModel = layer_model
        self.lineage: list[int] = layer_model.get_lineage(index)
        self.new_value: Any = value
        self.role: int = role
        self.old_value: Any
        if index.data(Roles.EditorRole) in ("integer", "float", "float_slider"):
            self.old_value = self.layer_model.data(index, role)[0] if old_value is None else old_value
        elif index.data(Roles.EditorRole) == "colormap":
            old_colormap = self.layer_model.data(index, Qt.ItemDataRole.DisplayRole)
            layer = self.layer_model.data(index, Qt.ItemDataRole.EditRole)
            old_v0, old_v1 = layer.colormap_v0, layer.colormap_v1
            self.old_value = (old_colormap, old_v0, old_v1)
        else:
            self.old_value = self.layer_model.data(index, role)

    def redo(self) -> None:
        index: QModelIndex = self.layer_model.get_index_from_lineage(self.lineage)
        if not index.isValid():
            raise RuntimeError("SetDataCommand: invalid lineage")
        if not self.layer_model.setData(index, self.new_value, self.role):
            self.setObsolete(True)

    def undo(self) -> None:
        index: QModelIndex = self.layer_model.get_index_from_lineage(self.lineage)
        if not index.isValid():
            raise RuntimeError("SetDataCommand: invalid lineage")
        self.layer_model.setData(index, self.old_value, self.role)


class AddLayerCommand(QUndoCommand):

    def __init__(self, layer_model: LayerModel, layer: Layer, index: int = -1,
                 parent_command: Optional[QUndoCommand] = None):
        super().__init__(parent_command)
        self.setText(f"Add layer {layer.name}")
        self.layer_model: LayerModel = layer_model
        self.layer: Layer = layer
        self.index: int = index

    def redo(self) -> None:
        if not self.layer_model.add_item(self.layer, self.index, QModelIndex()):
            self.setObsolete(True)
        else:
            self.layer.request_paint.connect(self.layer_model.request_paint)

    def undo(self) -> None:
        self.layer_model.remove_item(self.index, QModelIndex())
        self.layer.request_paint.disconnect(self.layer_model.request_paint)


class AddSelectionPointCommand(QUndoCommand):

    def __init__(self, layer_model: LayerModel, point: SelectionPoint,
                 parent_command: Optional[QUndoCommand] = None):
        super().__init__(parent_command)
        self.setText(f"Add point {point.name}")
        self.layer_model: LayerModel = layer_model
        self.point: SelectionPoint = point

    def redo(self) -> None:
        assert self.layer_model.selection is not None
        selection_index: QModelIndex = self.layer_model.index(
            self.layer_model.selection.child_number(), 0, QModelIndex())
        assert selection_index.isValid()
        points_folder: QModelIndex = self.layer_model.index(
            self.layer_model.selection.points_folder.child_number(), 0, selection_index)
        assert points_folder.isValid()
        if self.layer_model.add_item(self.point, -1, points_folder):
            self.layer_model.request_expand.emit(selection_index)
            self.layer_model.request_expand.emit(points_folder)
        else:
            self.setObsolete(True)

    def undo(self) -> None:
        assert self.layer_model.selection is not None
        selection_index: QModelIndex = self.layer_model.index(
            self.layer_model.selection.child_number(), 0, QModelIndex())
        assert selection_index.isValid()
        points_folder: QModelIndex = self.layer_model.index(
            self.layer_model.selection.points_folder.child_number(), 0, selection_index)
        assert points_folder.isValid()
        self.layer_model.remove_item(self.point.child_number(), points_folder)


class AddSelectionProfileCommand(QUndoCommand):

    def __init__(self, layer_model: LayerModel, profile: SelectionProfile,
                 parent_command: Optional[QUndoCommand] = None):
        super().__init__(parent_command)
        self.setText(f"Add profile {profile.name}")
        self.layer_model: LayerModel = layer_model
        self.profile: SelectionProfile = profile

    def redo(self) -> None:
        assert self.layer_model.selection is not None
        selection_index: QModelIndex = self.layer_model.index(
            self.layer_model.selection.child_number(), 0, QModelIndex())
        assert selection_index.isValid()
        profiles_folder: QModelIndex = self.layer_model.index(
            self.layer_model.selection.profiles_folder.child_number(), 0, selection_index)
        assert profiles_folder.isValid()
        if self.layer_model.add_item(self.profile, -1, profiles_folder):
            self.layer_model.request_expand.emit(selection_index)
            self.layer_model.request_expand.emit(profiles_folder)
        else:
            self.setObsolete(True)

    def undo(self) -> None:
        assert self.layer_model.selection is not None
        selection_index: QModelIndex = self.layer_model.index(
            self.layer_model.selection.child_number(), 0, QModelIndex())
        assert selection_index.isValid()
        profiles_folder: QModelIndex = self.layer_model.index(
            self.layer_model.selection.profiles_folder.child_number(), 0, selection_index)
        assert profiles_folder.isValid()
        self.layer_model.remove_item(self.profile.child_number(), profiles_folder)


class AddSelectionReferenceCommand(QUndoCommand):

    def __init__(self, layer_model: LayerModel, reference: SelectionReference,
                 parent_command: Optional[QUndoCommand] = None):
        super().__init__(parent_command)
        self.setText(f"Add reference {reference.name}")
        self.layer_model: LayerModel = layer_model
        self.reference: SelectionReference = reference

    def redo(self) -> None:
        assert self.layer_model.selection is not None
        selection_index: QModelIndex = self.layer_model.index(
            self.layer_model.selection.child_number(), 0, QModelIndex())
        assert selection_index.isValid()
        references_folder: QModelIndex = self.layer_model.index(
            self.layer_model.selection.references_folder.child_number(), 0, selection_index)
        assert references_folder.isValid()
        if self.layer_model.add_item(self.reference, -1, references_folder):
            self.layer_model.request_expand.emit(selection_index)
            self.layer_model.request_expand.emit(references_folder)
        else:
            self.setObsolete(True)

    def undo(self) -> None:
        assert self.layer_model.selection is not None
        selection_index: QModelIndex = self.layer_model.index(
            self.layer_model.selection.child_number(), 0, QModelIndex())
        assert selection_index.isValid()
        references_folder: QModelIndex = self.layer_model.index(
            self.layer_model.selection.references_folder.child_number(), 0, selection_index)
        assert references_folder.isValid()
        self.layer_model.remove_item(self.reference.child_number(), references_folder)


class RemoveTreeItemCommand(QUndoCommand):

    def __init__(self, layer_model: LayerModel, index: int, parent: QModelIndex,
                 parent_command: Optional[QUndoCommand] = None):
        super().__init__(parent_command)
        item_index: QModelIndex = layer_model.index(index, 0, parent)
        self.item: TreeItem = layer_model.get_item(item_index)
        self.setText(f"Remove {self.item.kind} {self.item.name}")
        self.layer_model: LayerModel = layer_model
        self.lineage: list[int] = layer_model.get_lineage(item_index)

    def redo(self) -> None:
        parent: QModelIndex = self.layer_model.get_index_from_lineage(self.lineage[:-1])
        if not self.layer_model.remove_item(self.lineage[-1], parent):
            self.setObsolete(True)

    def undo(self) -> None:
        parent: QModelIndex = self.layer_model.get_index_from_lineage(self.lineage[:-1])
        self.layer_model.add_item(self.item, self.lineage[-1], parent)
        while parent.isValid():
            self.layer_model.request_expand.emit(parent)
            parent = parent.parent()


class MoveLayerCommand(QUndoCommand):

    def __init__(self, layer_model: LayerModel, start: int, target: int,
                 parent_command: Optional[QUndoCommand] = None):
        super().__init__(parent_command)
        layer: TreeItem = layer_model.get_item(layer_model.index(start, 0, QModelIndex()))
        assert isinstance(layer, Layer)
        self.setText(f"Move layer {layer.name}")
        self.layer_model: LayerModel = layer_model
        self.start: int = start
        self.target: int = target
        if start == target:
            self.setObsolete(True)

    def redo(self) -> None:
        if not self.layer_model.move_item(self.start, self.target, QModelIndex()):
            self.setObsolete(True)

    def undo(self) -> None:
        if self.start < self.target:
            start: int = self.target - 1
            target: int = self.start
        elif self.start > self.target:
            start = self.target
            target = self.start + 1
        else:
            start = self.start
            target = self.start
        self.layer_model.move_item(start, target, QModelIndex())



# -*- coding: utf-8 -*-

"""
Basic matrix operations emulating OpenGL 3D transforms.

Copyright (c) 2010, Renaud Blanch <rndblnch at gmail dot com>
Licence: GPLv3 or higher <http://www.gnu.org/licenses/gpl.html>
"""


# imports ####################################################################

from math import (cos as _cos, sin as _sin,
                  radians as _radians, sqrt as _sqrt)

import itertools

_sum = sum

Matrix = list[list[float]]

# matrix construction ########################################################


def identity():
    return [[1., 0., 0., 0.],
            [0., 1., 0., 0.],
            [0., 0., 1., 0.],
            [0., 0., 0., 1.]]


def scale(sx=1., sy=1., sz=1.):
    return [[float(sx), 0.,        0.,        0.],
            [0.,        float(sy), 0.,        0.],
            [0.,        0.,        float(sz), 0.],
            [0.,        0.,        0.,        1.]]


def translate(tx=0., ty=0., tz=0.):
    return [[1., 0., 0., float(tx)],
            [0., 1., 0., float(ty)],
            [0., 0., 1., float(tz)],
            [0., 0., 0., 1.]]


def rotate(a=0., x=0., y=0., z=1.):
    a = _radians(a)
    s, c = _sin(a), _cos(a)
    h = _sqrt(x*x + y*y + z*z)
    x, y, z = x/h, y/h, z/h
    sx, sy, sz = s*x, s*y, s*z
    oc = 1.-c
    return [[oc*x*x+c,  oc*x*y-sz, oc*x*z+sy, 0.],
            [oc*x*y+sz, oc*y*y+c,  oc*y*z-sx, 0.],
            [oc*x*z-sy, oc*y*z+sx, oc*z*z+c,  0.],
            [0.,        0.,        0.,        1.]]


def ortho(l, r, b, t, n, f):
    w, h, d = r-l, t-b, f-n
    return [[2./w, 0.,    0.,   -(r+l)/w],
            [0.,   2./h,  0.,   -(t+b)/h],
            [0.,   0.,   -2./d, -(f+n)/d],
            [0.,   0.,    0.,    1.]]


def frustum(l, r, b, t, n, f):
    w, h, d = r-l, t-b, f-n
    return [[2.*n/w, 0.,      (r+l)/w,  0.],
            [0.,     2.*n/h,  (t+b)/h,  0.],
            [0.,     0.,     -(f+n)/d, -2.*f*n/d],
            [0.,     0.,     -1.,       0.]]


def from_rasterio_Affine(a):
    a = list(a)
    m = identity()
    m[0][0] = a[0]
    m[0][1] = a[1]
    m[0][3] = a[2]
    m[1][0] = a[3]
    m[1][1] = a[4]
    m[1][3] = a[5]
    return m

# manipulation ###############################################################


def matrix(n, m, f=lambda i, j: 0.):
    I, J = range(n), range(m)
    return [[float(f(i, j)) for j in J] for i in I]


def size(A):
    return len(A), len(A[0])


def transpose(A):
    n, m = size(A)
    return matrix(m, n, lambda i, j: A[j][i])


def exclude(A, i, j):
    return [R[:j]+R[j+1:] for R in A[:i]+A[i+1:]]


def top_left(A):
    n, m = size(A)
    return exclude(A, n-1, m-1)


def flatten(A):
    return list(itertools.chain.from_iterable(A))


# sum ########################################################################

def add(A, B):
    n, p = size(A)
    q, m = size(B)
    assert (n, p) == (q, m)
    return matrix(n, m, lambda i, j: A[i][j]+B[i][j])


def sub(A, B):
    n, p = size(A)
    q, m = size(B)
    assert (n, p) == (q, m)
    return matrix(n, m, lambda i, j: A[i][j]-B[i][j])


# product ####################################################################

def scalar(s, A):
    n, m = size(A)
    return matrix(n, m, lambda i, j: s*A[i][j])


def mul(A, B):
    n, p = size(A)
    q, m = size(B)
    assert p == q
    K = range(p)
    return matrix(n, m, lambda i, j: _sum(A[i][k]*B[k][j] for k in K))


def product(A, *Bs):
    for B in Bs:
        A = mul(A, B)
    return A


# inverse ####################################################################

def det(A):
    n, m = size(A)
    assert n == m
    if n == 1:
        return A[0][0]
    else:
        return sum(A[i][0]*cofactor(A, i, 0) for i in range(n))


def minor(A, i, j):
    return det(exclude(A, i, j))


def cofactor(A, i, j):
    return (-1)**(i+j)*minor(A, i, j)


def inverse(A):
    n, m = size(A)
    assert n == m
    C = matrix(n, m, lambda i, j: cofactor(A, i, j))
    I = range(n)
    d = sum(A[i][0]*C[i][0] for i in I)
    return scalar(1./d, transpose(C))



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
""" AbstractMapView

This module contains the class AbstractMapView that is a generic widget for
image views. It is inherited by MapView and MinimapView classes
(in the modules named MapView and MinimapView).
"""


# imports ###################################################################

import ctypes

from OpenGL import GL

import numpy as np

from PySide6.QtCore import Qt, Slot, Signal

from PySide6.QtWidgets import QWidget

from PySide6.QtOpenGLWidgets import QOpenGLWidget

from PySide6.QtGui import QMouseEvent, QWheelEvent

from PySide6.QtOpenGL import QOpenGLVertexArrayObject, QOpenGLBuffer

from shiboken6 import VoidPtr

from insarviz.map.MapModel import MapModel

from insarviz.linalg import matrix, vector

if True and False:
    import OpenGL
    OpenGL.ERROR_CHECKING = False
    OpenGL.ERROR_LOGGING = False
    OpenGL.ERROR_ON_COPY = True
    OpenGL.STORE_POINTERS = False

# Abstract map view ###############################################################


class AbstractMapView(QOpenGLWidget):
    """
    Abstract class for a map view, a QOpenGLWidget display of the
    displacement data, color coded using a colormap
    """

    INTERACTIVE: int = 0

    displayed_area_changed = Signal(tuple, tuple, tuple)

    def __init__(self, map_model: MapModel, parent: QWidget):
        """
        Generate AbstractView.

        Parameters
        ----------
        map_model : MapModel
        """
        super().__init__(parent)
        self.map_model: MapModel = map_model
        self.map_model.layer_model.dataChanged.connect(self.paint)
        self.map_model.layer_model.rowsInserted.connect(self.paint)
        self.map_model.layer_model.rowsMoved.connect(self.paint)
        self.map_model.layer_model.rowsRemoved.connect(self.paint)
        self.map_model.request_paint.connect(self.paint)
        self.map_model.request_set_view_center.connect(self.set_view_center)
        self.interaction: int = 0
        self.set_interaction(self.INTERACTIVE)
        self.left_drag: bool = False
        self.right_drag: bool = False
        # OpenGL vertex array object holding a square on which to display texture
        self.vao = QOpenGLVertexArrayObject()
        self.z: float = 1.0  # zoom level
        # center of the view matrix in data coordinates
        self.cx: float = 0.
        self.cy: float = 0.
        self.view_matrix: matrix.Matrix = matrix.identity()
        self.view_matrix_inverse: matrix.Matrix = matrix.identity()
        self.projection_matrix: matrix.Matrix = matrix.identity()
        # position of the last mouse press (xdata, ydata, xwidget, ywidget)
        self.p0: tuple[float, float, int, int] = 0., 0., 0, 0
        # last position of the mouse (xdata, ydata, xwidget, ywidget)
        self.p_prev: tuple[float, float, int, int] = 0., 0., 0, 0

    # opengl

    def initializeGL(self) -> None:
        glfunc = self.context().functions()
        glfunc.glClearColor(0.5, 0.5, 0.5, 1.)
        # build a Vertex Array Object that is a square from (0,0) to (1,1) mapped with a texture
        # also from (0,0) to (1,1). The vertex are given in such order that:
        # glDrawArrays(GL_LINE_LOOP, 0, 4) draws the square's border
        # glDrawElements(GL_TRIANGLES, 6, GL_UNSIGNED_INT, None) draws the textured square
        self.vao.create()
        self.vao.bind()
        vertex = np.ravel([
            # vertex_x, vertex_y, vertex_z, texture_x, texture_y
            [0., 0., 0., 0., 0.],
            [1., 0., 0., 1., 0.],
            [1., 1., 0., 1., 1.],
            [0., 1., 0., 0., 1.]
        ]).astype(np.float32)
        vertex_buffer = QOpenGLBuffer(QOpenGLBuffer.Type.VertexBuffer)
        vertex_buffer.create()
        vertex_buffer.setUsagePattern(QOpenGLBuffer.UsagePattern.StaticDraw)
        vertex_buffer.bind()
        vertex_buffer.allocate(bytes(vertex), vertex.nbytes)
        float_size: int = ctypes.sizeof(ctypes.c_float)
        vertex_offset = VoidPtr(0 * float_size)
        tex_coord_offset = VoidPtr(3 * float_size)
        record_len: int = 5 * float_size
        glfunc.glEnableVertexAttribArray(0)
        glfunc.glVertexAttribPointer(0, 3, int(GL.GL_FLOAT), int(GL.GL_FALSE), record_len,
                                     vertex_offset)
        glfunc.glVertexAttribPointer(1, 2, int(GL.GL_FLOAT), int(GL.GL_FALSE), record_len,
                                     tex_coord_offset)
        glfunc.glEnableVertexAttribArray(1)
        indices = np.array([0, 1, 3, 1, 2, 3], dtype='uint32')
        indices_buffer = QOpenGLBuffer(QOpenGLBuffer.Type.IndexBuffer)
        indices_buffer.create()
        indices_buffer.setUsagePattern(QOpenGLBuffer.UsagePattern.StaticDraw)
        indices_buffer.bind()
        indices_buffer.allocate(indices, indices.nbytes)
        self.vao.release()
        vertex_buffer.release()
        indices_buffer.release()

    def resizeGL(self, width: int, height: int) -> None:
        """
        Resize OpenGL view according to new settings

        Parameters
        ----------
        width : int
            new view width.
        height : int
            new view height.
        """
        glfunc = self.context().functions()
        glfunc.glViewport(0, 0, width, height)
        self.update_projection_matrix()
        self.update_view_matrix()

    # interaction
    @Slot(int)
    def set_interaction(self, value: int) -> None:
        self.interaction = value

    @Slot(float, float)
    def set_view_center(self, cx: float, cy: float) -> None:
        self.cx = cx
        self.cy = cy
        self.paint()

    def mousePressEvent(self, e: QMouseEvent) -> None:
        # pylint: disable=missing-function-docstring, invalid-name
        if self.map_model.current_band_index is not None and not self.left_drag and not self.right_drag:
            if e.button() == Qt.MouseButton.LeftButton:
                self.left_drag = True
            elif e.button() == Qt.MouseButton.RightButton:
                self.right_drag = True

    def mouseReleaseEvent(self, e: QMouseEvent) -> None:
        # pylint: disable=missing-function-docstring, invalid-name
        if self.left_drag and e.button() == Qt.MouseButton.LeftButton:
            self.left_drag = False
        elif self.right_drag and e.button() == Qt.MouseButton.RightButton:
            self.right_drag = False

    def wheelEvent(self, event: QWheelEvent) -> None:
        # pylint: disable=invalid-name
        """
        Update zoom value according to mouse wheel angle change
        """
        x: float
        y: float
        x, y = self.get_texture_coordinate(int(event.position().x()), int(event.position().y()))
        ds: float = event.angleDelta().y()/8  # degrees
        self.zoom(ds, x, y)

    @Slot(float, float, float)
    def zoom(self, ds: float, x: float, y: float) -> None:
        """
        Called by MouseMoveEvent and wheelEvent on Map. Update Map variables
        for display at new zoom level.

        Parameters
        ----------
        ds : float
            zoom change (mouse wheel angle in degrees or right-click+drag distance).
        x : float
            x-axis position of the zoom change focal point.
        y : float
            y-axis position of the zoom change focal point.
        """
        dz: float = np.exp(ds*0.01)
        self.z *= dz
        # for (x,y) to stay at the same position in the view
        # the following equations must be satistied:
        # (x - cx_i) * z = (x - cx_i+1) * z * dz
        # (y - cy_i) * z = (y - cy_i+1) * z * dz
        # (distance of (x,y) from the center of the view shall remain the same after zooming)
        self.cx = x - (x - self.cx) / dz
        self.cy = y - (y - self.cy) / dz
        self.update_view_matrix()
        self.update()

    @Slot(float, float)
    def pan(self, dx: float, dy: float) -> None:
        """
        Move the camera (translation in x/y)
        Parameters
        ----------
        dx : float
            change in Map view position along x-axis.
        dy : float
            change in Map view position along y-axis.
        """
        self.cx += dx
        self.cy += dy
        self.update_view_matrix()
        self.update()

    def get_texture_coordinate(self, x: int, y: int) -> tuple[float, float]:
        """
        Transform screen coord to texture coord.

        Parameters
        ----------
        x : int
        y : int

        Returns
        -------
        tuple (float, float)
            texture coordinates
        """
        screen_coord = vector.matrix((float(x), float(y), 0., 1.))
        texture_coord = matrix.product(self.view_matrix_inverse, screen_coord)
        return texture_coord[0][0], texture_coord[1][0]

    def update_view_matrix(self) -> None:
        flip_matrix = matrix.identity()
        if self.map_model.flip_h:
            flip_matrix = matrix.mul(matrix.scale(sx=-1.), flip_matrix)
            flip_matrix = matrix.mul(matrix.translate(tx=self.width()), flip_matrix)
        if self.map_model.flip_v:
            flip_matrix = matrix.mul(matrix.scale(sy=-1.), flip_matrix)
            flip_matrix = matrix.mul(matrix.translate(ty=self.height()), flip_matrix)
        self.view_matrix = matrix.product(
            flip_matrix,
            matrix.translate(self.width()//2, self.height()//2),
            matrix.scale(self.z, self.z),
            matrix.translate(-self.cx, -self.cy)
        )
        self.view_matrix_inverse = matrix.product(
            matrix.translate(self.cx, self.cy),
            matrix.scale(1./self.z, 1./self.z),
            matrix.translate(-self.width()//2, -self.height()//2),
            flip_matrix
        )
        self.displayed_area_changed.emit(self.get_texture_coordinate(0, 0),
                                         self.get_texture_coordinate(self.width(), self.height()),
                                         (self.width(), self.height()))

    def update_projection_matrix(self) -> None:
        # vertical case is reversed because OpenGl Y axis is flipped with respect to data Y axis
        self.projection_matrix = matrix.ortho(0, self.width(), self.height(), 0, -1, 1)

    @Slot()
    def paint(self) -> None:
        self.update_view_matrix()
        self.update_projection_matrix()
        self.update()



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
""" MapView

This module handles the creation of the Map view (image display of a band data)
 and user interactions with it (pan, zoom, data selection...).
 Works with the module MapModel (as in a Model/View architecture).

Contains class:
    * MapView
"""


# imports ###################################################################

from typing import Optional
import rasterio

from OpenGL import GL

from PySide6.QtCore import QSize, Slot, Signal, Qt, QPoint, QPointF, QEvent

from PySide6.QtWidgets import QToolTip, QWidget, QMenu, QMessageBox

from PySide6.QtGui import (
    QPainter, QMouseEvent, QIcon, QCursor, QColor, QAction, QActionGroup, QGuiApplication
)

from insarviz.map.AbstractMapView import AbstractMapView

from insarviz.map.MapModel import MapModel

from insarviz.linalg import matrix

from insarviz.map.layers.SelectionLayer import (
    SelectionItem, SelectionPoint, SelectionProfile, SelectionReference
)
from insarviz.map.layers.Layer import Layer

# MapView #######################################################################

class MapView(AbstractMapView):

    size_threshold_confirmation_selection_item = 1000000

    POINTS: int = 1
    PROFILE: int = 2
    REF: int = 3

    # send the position of the mouse in data coordinate (int, int) and the band number to PlotModel
    pointer_changed = Signal(tuple, int)
    viewport_matrix_changed = Signal(object)
    interaction_changed = Signal(int)
    request_lock_axes = Signal()

    def __init__(self, map_model: MapModel, parent: QWidget):
        """
        Generate Map, a QOpenGLWidget display of the displacement data. Zoom
        level can be interactively set through mouse wheel or right-click+drag.
        View position on the Map can be interactively set through left-click+
        drag. View is synchronized to Minimap.

        Parameters
        ----------
        map_model : MapModel
            Model managing the data for Map and Minimap.
        """
        super().__init__(map_model, parent)
        # selection item that the user is creating before adding it to the selection layer
        self.selection_item: Optional[SelectionItem] = None
        self.setCursor(Qt.CursorShape.OpenHandCursor)
        self.menu = QMenu()
        self.action_group = QActionGroup(self)
        self.lock_plot_axes_action = QAction("Lock axes", self)
        self.lock_plot_axes_action.triggered.connect(self.lock_plot_axes)
        self.menu.addAction(self.lock_plot_axes_action)
        self.action_group.addAction(self.lock_plot_axes_action)

        self.create_temp_point_action = QAction("Create temporary point", self)
        self.create_temp_point_action.setIcon(QIcon('icons:points.png'))
        self.create_temp_point_action.triggered.connect(self.create_temp_point)
        self.menu.addAction(self.create_temp_point_action)
        self.action_group.addAction(self.create_temp_point_action)

        self.create_point_action = QAction("Create point", self)
        self.create_point_action.setIcon(QIcon('icons:points.png'))
        self.create_point_action.triggered.connect(self.create_point)
        self.menu.addAction(self.create_point_action)
        self.action_group.addAction(self.create_point_action)
        self.create_profile_action = QAction("Create profile", self)
        self.create_profile_action.setIcon(QIcon('icons:profile.png'))
        self.create_profile_action.triggered.connect(self.create_profile)
        self.menu.addAction(self.create_profile_action)
        self.action_group.addAction(self.create_profile_action)
        self.create_reference_action = QAction("Create reference", self)
        self.create_reference_action.setIcon(QIcon('icons:ref.png'))
        self.create_reference_action.triggered.connect(self.create_reference)
        self.menu.addAction(self.create_reference_action)
        self.action_group.addAction(self.create_reference_action)
        self.copy_coordinates_action = QAction("Copy coordinates to clipboard", self)
        self.copy_coordinates_action.triggered.connect(self.copy_coordinates_to_clipboard)
        self.menu.addAction(self.copy_coordinates_action)
        self.action_group.addAction(self.copy_coordinates_action)

    # opengl

    def sizeHint(self) -> QSize:
        return QSize(300, 300)

    def paintGL(self) -> None:
        glfunc = self.context().functions()
        painter = QPainter(self)
        # FIX for issue #150
        # Qpainter seems to need to "paint" something in order for it to correctly give back the
        # global OpenGL state (I can't find why)
        painter.fillRect(0, 0, 0, 0, QColor("transparent"))
        # END FIX
        painter.beginNativePainting()
        glfunc.glClear(GL.GL_COLOR_BUFFER_BIT)
        painter.endNativePainting()
        # layers is reversed so first layer is displayed last (i.e. upon the others)

        show_params = Layer.ShowParams()
        for layer in reversed(self.map_model.layer_model.layers):
            if layer.visible:
                layer.show(self.view_matrix, self.projection_matrix, show_params,
                           painter=painter, vao=self.vao, glfunc=glfunc)
                layer.update_show_params(show_params)
        if self.selection_item is not None:
            self.selection_item.show(self.view_matrix, painter, preview=True)
        painter.end()

    def mousePressEvent(self, e: QMouseEvent) -> None:
        if (self.map_model.current_band_index is not None and (not self.left_drag) and (not self.right_drag)
                and self.selection_item is None):
            self.pointer_changed.emit((), None)
            if e.button() in (Qt.MouseButton.LeftButton, Qt.MouseButton.RightButton):
                self.p0 = (*self.get_texture_coordinate(int(e.position().x()), int(e.position().y())),
                           int(e.position().x()), int(e.position().y()))
            if e.button() == Qt.MouseButton.LeftButton:
                if self.interaction == self.INTERACTIVE:
                    self.setCursor(Qt.CursorShape.ClosedHandCursor)
                else:
                    # check if pointer inside dataset
                    if ((0 <= self.p0[0] < self.map_model.tex_width)
                            and (0 <= self.p0[1] < self.map_model.tex_height)):
                        if self.interaction == self.POINTS:
                            self.create_point()
                        elif self.interaction == self.PROFILE:
                            self.create_profile()
                        elif self.interaction == self.REF:
                            self.create_reference()
            elif e.button() == Qt.MouseButton.RightButton:
                self.show_context_menu()
        super().mousePressEvent(e)
        # prevent right drag because right press launch context menu thus right release is missed
        self.right_drag = False

    def mouseMoveEvent(self, e: QMouseEvent) -> None:
        # check if data loaded (needed when opening without data)
        if self.map_model.current_band_index is not None:
            if (not self.left_drag) and (not self.right_drag):
                self.p0 = (*self.get_texture_coordinate(int(e.position().x()), int(e.position().y())),
                           int(e.position().x()), int(e.position().y()))
                i, j = int(self.p0[0]), int(self.p0[1])
                # check if pointer inside map
                if (0 <= i < self.map_model.tex_width) and (0 <= j < self.map_model.tex_height):
                    self.pointer_changed.emit((i, j), self.map_model.current_band_index)
                else:
                    self.pointer_changed.emit((), None)
            if self.interaction == self.INTERACTIVE:
                if self.left_drag:
                    # get the data coordinates of the point we are dragging
                    x0, y0, _, _ = self.p0
                    # get the data coordinates of where the cursor no is
                    x1, y1 = self.get_texture_coordinate(int(e.position().x()),
                                                         int(e.position().y()))
                    # make the difference between both and move the center of the view accordingly
                    dx, dy = x1-x0, y1-y0
                    self.pan(-dx, -dy)
            elif self.interaction == self.POINTS and self.left_drag:
                if self.selection_item is not None:
                    assert isinstance(self.selection_item, SelectionPoint)
                    x, y = self.get_texture_coordinate(int(e.position().x()),
                                                       int(e.position().y()))
                    dx = abs(int(x) - self.selection_item.x) + 1
                    dy = abs(int(y) - self.selection_item.y) + 1
                    r = min(max(dx, dy), self.selection_item.x+1, self.selection_item.y+1,
                            self.map_model.tex_width - self.selection_item.x,
                            self.map_model.tex_height - self.selection_item.y)
                    if self.selection_item.r != r:
                        self.selection_item.set_r(r)
                        self.paint()
            elif self.interaction == self.PROFILE:
                if self.selection_item is not None:
                    assert isinstance(self.selection_item, SelectionProfile)
                    x, y = self.get_texture_coordinate(int(e.position().x()),
                                                       int(e.position().y()))
                    x, y = int(x), int(y)
                    if (0 <= x < self.map_model.tex_width and 0 <= y < self.map_model.tex_height):
                        if (x != self.selection_item.points[-1][0] or
                                y != self.selection_item.points[-1][1]):
                            if len(self.selection_item.points) > 1:
                                self.selection_item.remove_last_point()
                            self.selection_item.add_point(x, y)
                            self.paint()
            elif self.interaction == self.REF and self.left_drag:
                if self.selection_item is not None:
                    assert isinstance(self.selection_item, SelectionReference)
                    x, y = self.get_texture_coordinate(int(e.position().x()),
                                                       int(e.position().y()))
                    x, y = int(x), int(y)
                    x = max(x, 0)
                    x = min(x, self.map_model.tex_width - 1)
                    y = max(y, 0)
                    y = min(y, self.map_model.tex_height - 1)
                    x0, y0 = int(self.p0[0]), int(self.p0[1])
                    left, right = min(x, x0), max(x, x0)
                    top, bottom = min(y, y0), max(y, y0)
                    if self.selection_item.get_rect() != (left, top, right, bottom):
                        self.selection_item.set_rect(left, top, right, bottom)
                        self.paint()

    def mouseReleaseEvent(self, e: QMouseEvent) -> None:
        if self.interaction == self.INTERACTIVE and e.button() == Qt.MouseButton.LeftButton and self.left_drag:
            self.setCursor(Qt.CursorShape.OpenHandCursor)
        elif self.interaction == self.POINTS and e.button() == Qt.MouseButton.LeftButton:
            if self.selection_item is not None:
                assert isinstance(self.selection_item, SelectionPoint)
                size = (2*self.selection_item.r-1)**2
                if size > self.size_threshold_confirmation_selection_item:
                    if QMessageBox.question(self, "Insarviz",
                                            f"You are about to add a point of size {size}, it can "
                                            "take some time to compute its mean.\nAre you sure ?",
                                            QMessageBox.StandardButton.Cancel | QMessageBox.StandardButton.Ok,
                                            QMessageBox.StandardButton.Ok) == QMessageBox.StandardButton.Cancel:
                        self.selection_item = None
                        self.paint()
                        super().mouseReleaseEvent(e)
                        return None
                self.map_model.layer_model.add_selection_point(self.selection_item)
                self.selection_item = None
            self.set_interaction(self.INTERACTIVE)
        elif self.interaction == self.PROFILE and e.button() == Qt.MouseButton.LeftButton:
            if self.selection_item is not None:
                assert isinstance(self.selection_item, SelectionProfile)
                if len(self.selection_item.points) > 1:
                    self.selection_item.remove_last_point()
                    x, y = self.get_texture_coordinate(int(e.position().x()),
                                                       int(e.position().y()))
                    x, y = int(x), int(y)
                    if (0 <= x < self.map_model.tex_width and 0 <= y < self.map_model.tex_height):
                        self.selection_item.add_point(x, y)
                        self.selection_item.add_point(x, y)
                        self.paint()
        elif self.interaction == self.PROFILE and e.button() == Qt.MouseButton.RightButton:
            if self.selection_item is not None:
                assert isinstance(self.selection_item, SelectionProfile)
                self.selection_item.remove_last_point()
                if len(self.selection_item.points) > 1:
                    self.map_model.layer_model.add_selection_profile(self.selection_item)
                self.selection_item = None
                self.paint()
            self.set_interaction(self.INTERACTIVE)
        elif self.interaction == self.REF and e.button() == Qt.MouseButton.LeftButton:
            if self.selection_item is not None:
                assert isinstance(self.selection_item, SelectionReference)
                ref = self.selection_item
                size = (ref.right - ref.left + 1) * (ref.bottom - ref.top + 1)
                if size > self.size_threshold_confirmation_selection_item:
                    if QMessageBox.question(self, "Insarviz",
                                            f"You are about to add a reference of size {size}, it "
                                            "can take some time to compute its mean.\n"
                                            "Are you sure ?",
                                            QMessageBox.StandardButton.Cancel | QMessageBox.StandardButton.Ok,
                                            QMessageBox.StandardButton.Ok) == QMessageBox.StandardButton.Cancel:
                        self.selection_item = None
                        self.paint()
                        super().mouseReleaseEvent(e)
                        return None
                self.map_model.layer_model.add_selection_reference(self.selection_item)
                self.selection_item = None
            self.set_interaction(self.INTERACTIVE)
        super().mouseReleaseEvent(e)

    def leaveEvent(self, event: QEvent) -> None:
        super().leaveEvent(event)
        self.pointer_changed.emit((), None)

    def update_view_matrix(self) -> None:
        super().update_view_matrix()
        viewport_matrix = matrix.mul(self.view_matrix_inverse,
                                     matrix.scale(self.width(), self.height()))
        self.viewport_matrix_changed.emit(viewport_matrix)

    def show_context_menu(self):
        if ((0 <= self.p0[0] < self.map_model.tex_width)
                and (0 <= self.p0[1] < self.map_model.tex_height)):
            self.action_group.setEnabled(True)
        else:
            self.action_group.setDisabled(True)
        action = self.menu.exec(self.mapToGlobal(QPoint(self.p0[2], self.p0[3])))
        if action in (self.create_point_action, self.create_reference_action):
            # start left dragging to edit the newly created selection item
            # not required for profile because its edition does not use dragging
            self.left_drag = True
        # raise a mouseMoveEvent to update the position of the mouse after clicking on the menu
        event = QMouseEvent(QEvent.MouseMove, QPointF(self.mapFromGlobal(QCursor.pos())),
                            Qt.NoButton, Qt.NoButton, Qt.NoModifier)
        self.mouseMoveEvent(event)

    @Slot()
    def copy_coordinates_to_clipboard(self):
        _,_,x_widget,y_widget = self.p0
        dataset = self.map_model.loader.dataset
        if dataset.crs is not None:
            x,y = dataset.xy(y_widget,x_widget)
            lon, lat = rasterio.warp.transform(dataset.crs,
                                               rasterio.crs.CRS.from_epsg(4326),
                                               [x], [y])
            QGuiApplication.clipboard().setText(f"{lat[0]},{lon[0]}")

    @Slot()
    def lock_plot_axes(self):
        i, j = int(self.p0[0]), int(self.p0[1])
        assert (0 <= i < self.map_model.tex_width) and (0 <= j < self.map_model.tex_height)
        self.pointer_changed.emit((i, j), self.map_model.current_band_index)
        self.request_lock_axes.emit()

    @Slot()
    def create_point(self):
        i, j = int(self.p0[0]), int(self.p0[1])
        assert (0 <= i < self.map_model.tex_width) and (0 <= j < self.map_model.tex_height)
        self.set_interaction(self.POINTS)
        self.selection_item = SelectionPoint(i, j, 1)
        self.paint()

    @Slot()
    def create_temp_point(self):
        i, j = int(self.p0[0]), int(self.p0[1])
        assert (0 <= i < self.map_model.tex_width) and (0 <= j < self.map_model.tex_height)
        selection_item = SelectionPoint(i, j, 1)
        selection_item.color = QColor('red')
        selection_item.name = "Temporary Point"
        self.map_model.layer_model.add_selection_point(selection_item)

    @Slot()
    def create_reference(self):
        i, j = int(self.p0[0]), int(self.p0[1])
        assert (0 <= i < self.map_model.tex_width) and (0 <= j < self.map_model.tex_height)
        self.set_interaction(self.REF)
        self.selection_item = SelectionReference(i, j)
        self.paint()

    @Slot()
    def create_profile(self):
        i, j = int(self.p0[0]), int(self.p0[1])
        assert (0 <= i < self.map_model.tex_width) and (0 <= j < self.map_model.tex_height)
        self.set_interaction(self.PROFILE)
        self.selection_item = SelectionProfile(i, j, 1)
        self.paint()

    @Slot(tuple)
    def update_mouse_tooltip(self, info: tuple) -> None:
        if info == ():
            QToolTip.hideText()
        else:
            x, y, _, value = info
            x_data, y_data, x_widget, y_widget = self.p0
            x_data, y_data = int(x_data), int(y_data)
            if (not x == x_data) or (not y == y_data):
                print("pointer info is too late")
                return None
            p = self.mapToGlobal(QPoint(x_widget, y_widget))
            # force the tooltip to update its position even if text remain the same
            # QToolTip.showText(p, "")
            QToolTip.showText(p, f"x:{x}\ny:{y}\nval:{value:.3f}")

    # interaction
    @Slot(int)
    def set_interaction(self, value: int) -> None:
        if value != self.interaction:
            self.selection_item = None
            if value == self.INTERACTIVE:
                self.setCursor(Qt.CursorShape.OpenHandCursor)
            else:
                self.setCursor(Qt.CursorShape.CrossCursor)
            self.paint()
            super().set_interaction(value)
            self.interaction_changed.emit(value)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Adapted from PyQt Editable Tree Model Example
https://doc.qt.io/qtforpython-6/examples/example_widgets_itemviews_editabletreemodel.html
"""

from __future__ import annotations

from typing import Any, Optional, Union, overload

import weakref

from PySide6.QtCore import QObject, QModelIndex, Qt, QAbstractItemModel


# TreeItem class ###################################################################################

class TreeItem(QObject):

    kind: str = "item"  # description of the class

    def __init__(self, parent: Optional[TreeItem] = None):
        super().__init__()
        self.name: str = ""
        self._parent_item: Optional[weakref.ReferenceType] = None
        self.parent_item = parent
        self.child_items: list[TreeItem] = []

    @property
    def parent_item(self) -> Optional[TreeItem]:
        if self._parent_item is None:
            return None
        return self._parent_item()

    @parent_item.setter
    def parent_item(self, item: Optional[TreeItem]):
        # use weakref to prevent circular references between item and its parent that would prevent
        # garbage collection (we want to delete the textures of layers for example)
        if item is not None:
            self._parent_item = weakref.ref(item)

    def parent(self) -> Optional[TreeItem]:
        return self.parent_item

    def child(self, number: int) -> Optional[TreeItem]:
        if number < 0 or number >= len(self.child_items):
            return None
        return self.child_items[number]

    def last_child(self) -> Optional[TreeItem]:
        return self.child_items[-1] if self.child_items else None

    def child_count(self) -> int:
        return len(self.child_items)

    def child_number(self) -> int:
        if self.parent_item is not None:
            return self.parent_item.child_items.index(self)
        return 0

    def data(self, column: int, role: int) -> Any:
        raise NotImplementedError

    def flags(self, flags: Qt.ItemFlags, column: int) -> Qt.ItemFlags:
        raise NotImplementedError

    def add_child(self, child: TreeItem, position: int = -1) -> bool:
        if position > len(self.child_items) or position < -1:
            return False
        elif position == -1:
            position = len(self.child_items)
        child.parent_item = self
        self.child_items.insert(position, child)
        return True

    def remove_children(self, position: int, count: int) -> bool:
        for _ in range(count):
            self.child_items.pop(position)
        return True

    def set_data(self, value, column: int, role: int) -> bool:
        raise NotImplementedError


# TreeModel class ##################################################################################

class TreeModel(QAbstractItemModel):
    # methods of the form methodName are inherited from QAbstractItemModel
    # methods of the form method_name are custom methods

    # WARNING removeRows(self, row, count, parent) SHOULD NOT BE DEFINED
    # (Qt tries to remove dragged rows but here they are just moved in the child_items list)

    # index of the remove column (column indexes start at 0, so the second column has index 1)
    remove_column: int = 1

    def __init__(self, parent=None):
        super().__init__(parent)
        self.root_item = TreeItem()

    def columnCount(self, parent: QModelIndex = QModelIndex()) -> int:
        raise NotImplementedError

    def data(self, index: QModelIndex, role: int = Qt.ItemDataRole.DisplayRole):
        if not index.isValid():
            return None
        item: TreeItem = self.get_item(index)
        return item.data(index.column(), role)

    def flags(self, index: QModelIndex) -> Qt.ItemFlags:
        if not index.isValid():
            return Qt.ItemFlag.NoItemFlags | Qt.ItemFlag.ItemIsDropEnabled
        item: TreeItem = self.get_item(index)
        return item.flags(QAbstractItemModel.flags(self, index), index.column())

    def get_item(self, index: QModelIndex = QModelIndex()) -> TreeItem:
        if index.isValid():
            item: TreeItem = index.internalPointer()
            if item:
                return item
        return self.root_item

    def index(self, row: int, column: int, parent: QModelIndex = QModelIndex()) -> QModelIndex:
        if parent.isValid() and parent.column() != 0:
            return QModelIndex()
        parent_item: TreeItem = self.get_item(parent)
        if not parent_item:
            return QModelIndex()
        item: Optional[TreeItem] = parent_item.child(row)
        if item is not None:
            return self.createIndex(row, column, item)
        return QModelIndex()

    # see https://mypy.readthedocs.io/en/stable/more_types.html#function-overloading
    @overload
    def parent(self, child: QModelIndex) -> QModelIndex: ...

    @overload
    def parent(self) -> QObject: ...

    def parent(self, child: Optional[QModelIndex] = None) -> Union[QObject, QModelIndex]:
        if child is None:
            return super().parent()
        if not child.isValid():
            return QModelIndex()
        item: TreeItem = self.get_item(child)
        parent_item: Optional[TreeItem]
        if item:
            parent_item = item.parent()
        else:
            parent_item = None
        if parent_item == self.root_item or parent_item is None:
            return QModelIndex()
        return self.createIndex(parent_item.child_number(), 0, parent_item)

    def rowCount(self, parent: QModelIndex = QModelIndex()) -> int:
        if parent.isValid() and parent.column() > 0:
            return 0
        parent_item: TreeItem = self.get_item(parent)
        if not parent_item:
            return 0
        return parent_item.child_count()

    def setData(self, index: QModelIndex, value: Any, role: int = Qt.ItemDataRole.EditRole) -> bool:
        item: TreeItem = self.get_item(index)
        result: bool = item.set_data(value, index.column(), role)
        if result:
            self.dataChanged.emit(index, index, [role])
        return result

    def removeRows(self, row: int, count: int, parent: QModelIndex = QModelIndex()) -> bool:
        # DO NOT SUBCLASS
        return False

    # custom methods

    def add_item(self, item: 'TreeItem', row: int, parent: QModelIndex = QModelIndex()) -> bool:
        parent_item: TreeItem = self.get_item(parent)
        if not parent_item:
            return False
        if row < -1 or row > self.rowCount(parent):
            return False
        if row == -1:
            row = self.rowCount(parent)
        self.beginInsertRows(parent, row, row)
        parent_item.add_child(item, row)
        self.endInsertRows()
        return True

    def move_item(self, start: int, target: int, parent: QModelIndex) -> bool:
        parent_item: TreeItem = self.get_item(parent)
        if not parent_item:
            return False
        if start == target:
            return True
        if target == parent_item.child_count() and start == target-1:
            # no need to move to the last row as it is already the last one
            return True
        self.beginMoveRows(parent, start, start, parent, target)
        if start < target:
            # item is removed before being added back, so if start < target that will
            # reduce the new position of target
            target -= 1
        item: TreeItem = parent_item.child_items.pop(start)
        parent_item.child_items.insert(target, item)
        self.endMoveRows()
        return True

    def remove_item(self, position: int, parent: QModelIndex = QModelIndex()) -> bool:
        parent_item: TreeItem = self.get_item(parent)
        if not parent_item:
            return False
        if position < -1 or position + 1 > self.rowCount(parent):
            return False
        if position == -1:
            position = self.rowCount(parent) - 1
        self.beginRemoveRows(parent, position, position)
        parent_item.remove_children(position, 1)
        self.endRemoveRows()
        return True

    def clear(self) -> None:
        if self.rowCount():
            self.beginResetModel()
            self.root_item.remove_children(0, self.rowCount())
            self.endResetModel()



# -*- coding: utf-8 -*-

""" PaletteView

This module handles the creation of the Palette view: an interactive plot
of the distribution of data values and its corresponding color scheme
Also handles user interactions with it (change color gradient and its
correspondance with data values)

Contains class:
    * Palette
"""

# imports ##########################################################################################

from typing import Optional

import logging

import numpy as np

from PySide6.QtCore import Qt, Signal, Slot, QSize, QEvent

from PySide6.QtWidgets import QWidget, QVBoxLayout, QToolBar, QPushButton

from PySide6.QtGui import QResizeEvent, QColor

from pyqtgraph import HistogramLUTWidget, BarGraphItem, LegendItem, GraphicsView
from pyqtgraph.widgets.ColorMapMenu import ColorMapMenu, buildMenuEntryAction, PrivateActionData

from pyqtgraph.colormap import ColorMap

from insarviz.colormaps import my_colormaps, my_cyclic_colormaps

from insarviz.custom_widgets import QtWaitingSpinner


logger = logging.getLogger(__name__)


# ColormapWidget class #############################################################################

class ColormapWidget(QWidget):

    autorange_threshold = 0.02
    default_padding = 0.15
    max_padding = 0.4

    # connected to MapModel.set_v0_v1
    v0_v1_changed = Signal(float, float)
    # connected to MapModel.set_colormap
    colormap_changed = Signal(ColorMap)
    #
    compute_histograms = Signal(tuple)
    cancel_compute_histograms = Signal()

    def __init__(self, parent: Optional[QWidget] = None):
        super().__init__(parent=parent)
        self.histogram_widget = myHistogramWidget(self)
        # add a curve for total histogram
        self.total_hist_plot = BarGraphItem(x0=[], x1=[], y0=[], y1=[],
                                            pen=None, brush=(20, 20, 220, 100))
        self.total_hist_plot.setZValue(10)
        self.total_hist_plot.setRotation(90)
        self.histogram_widget.vb.addItem(self.total_hist_plot)
        # add a curve for band histogram
        # (total data histogram use the base HistogramLUTWidget mono curve: self.plot)
        self.band_hist_plot = BarGraphItem(x0=[], x1=[], y0=[], y1=[],
                                           pen=None, brush=(220, 20, 20, 100))
        self.band_hist_plot.setZValue(10)
        self.band_hist_plot.setRotation(90)
        self.histogram_widget.vb.addItem(self.band_hist_plot)
        # legend
        self.legend = LegendItem()
        self.legend.addItem(self.total_hist_plot, "all bands")
        self.legend.addItem(self.band_hist_plot, "current band")
        self.legend_widget = myLegendWidget(parent=parent)
        self.legend_widget.setCentralWidget(self.legend)
        # toolbar
        self.toolbar = QToolBar(self)
        self.toolbar.setOrientation(Qt.Orientation.Vertical)
        self.autorange_total_button: QPushButton = QPushButton("Autorange all bands", self)
        self.autorange_total_button.setToolTip("Automatically adjust the colormap mapping using the"
                                               " all band histogram")
        self.autorange_total_button.clicked.connect(self.autorange_total)
        self.toolbar.addWidget(self.autorange_total_button)
        self.autorange_band_button: QPushButton = QPushButton("Autorange current band", self)
        self.autorange_band_button.setToolTip("Automatically adjust the colormap mapping using the"
                                              " current band histogram")
        self.autorange_band_button.clicked.connect(self.autorange_band)
        self.toolbar.addWidget(self.autorange_band_button)
        self.compute_histogram_button: QPushButton = QPushButton("Recompute histograms", self)
        self.compute_histogram_button.setToolTip(
            "Recompute the histograms using the area to define outliers")
        self.compute_histogram_button.clicked.connect(self.request_histograms)
        self.compute_action = self.toolbar.addWidget(self.compute_histogram_button)
        self.cancel_button: QPushButton = QPushButton("Cancel", self)
        self.cancel_button.setToolTip("Cancel histogram computation")
        self.cancel_button.clicked.connect(self.cancel_compute_histograms)
        self.cancel_action = self.toolbar.addWidget(self.cancel_button)
        # layout
        self.main_layout = QVBoxLayout(self)
        self.main_layout.addWidget(self.toolbar)
        self.main_layout.addWidget(self.histogram_widget, stretch=1)
        self.main_layout.addWidget(self.legend_widget)
        self.setLayout(self.main_layout)
        # internal signal and slots
        self.histogram_widget.sigLevelsChanged.connect(
            lambda _: self.v0_v1_changed.emit(*self.histogram_widget.getLevels()))
        self.histogram_widget.gradient.menu.sigColorMapTriggered.connect(
            self.colormap_changed.emit)
        # set disabled
        self.setDisabled(True)
        self.cancel_action.setVisible(False)
        self.spinner = QtWaitingSpinner(self.histogram_widget)

    @Slot()
    def on_compute_histogram(self) -> None:
        self.histogram_widget.setDisabled(True)
        self.legend_widget.setDisabled(True)
        self.autorange_total_button.setDisabled(True)
        self.autorange_band_button.setDisabled(True)
        self.compute_histogram_button.setDisabled(True)
        if not len(self.total_hist_plot.opts.get('x0')) == 0:
            self.compute_action.setVisible(False)
            self.cancel_action.setVisible(True)
        self.spinner.start()

    @Slot()
    def request_histograms(self) -> None:
        self.compute_histograms.emit(self.histogram_widget.getLevels())

    @Slot()
    def on_histograms_computed(self) -> None:
        self.setEnabled(True)
        self.histogram_widget.setEnabled(True)
        self.legend_widget.setEnabled(True)
        self.autorange_total_button.setEnabled(True)
        self.autorange_band_button.setEnabled(True)
        self.compute_histogram_button.setEnabled(True)
        self.compute_action.setVisible(True)
        self.cancel_action.setVisible(False)
        self.spinner.stop()

    def resizeEvent(self, event: QResizeEvent) -> None:
        super().resizeEvent(event)
        width = 0.85*self.histogram_widget.size().width()
        self.spinner.setInnerRadius(4*width//20)
        self.spinner.setLineLength(6*width//20)
        self.spinner.setLineWidth(max(2, (2*width)//20))

    def sizeHint(self) -> QSize:
        return QSize(self.minimumWidth(), self.minimumHeight())

    def on_close(self) -> None:
        self.set_total_histogram(np.array([]), np.array([]))
        self.set_band_histogram(np.array([]), np.array([]))
        self.set_default_colormap()
        self.setDisabled(True)

    @Slot(np.ndarray, np.ndarray)
    def set_total_histogram(self, hist: np.ndarray, bins: np.ndarray) -> None:
        self.total_hist_plot.setOpts(x0=bins[:-1], x1=bins[1:], y0=np.zeros(len(hist)), y1=hist)
        if len(hist) == 0:
            return
        self.autorange_total()
        self.histogram_widget.vb.setYRange(bins[1], bins[-2], padding=self.default_padding)
        ymin = float(bins[0] - (bins[-1] - bins[0]) * self.max_padding)
        ymax = float(bins[-1] + (bins[-1] - bins[0]) * self.max_padding)
        self.histogram_widget.vb.setLimits(yMin=ymin, yMax=ymax)

    @Slot(np.ndarray, np.ndarray)
    def set_band_histogram(self, hist: np.ndarray, bins: np.ndarray) -> None:
        self.band_hist_plot.setOpts(x0=bins[:-1], x1=bins[1:], y0=np.zeros(len(hist)), y1=hist)

    def autorange_from_hist(self, hist: np.ndarray, bins: np.ndarray) -> tuple[float, float]:
        assert len(hist) > 2 and len(hist)+1 == len(bins)
        # get the first index where the histogram cumulative sum is greater than threshold
        idx_v0 = np.where(np.cumsum(hist) > self.autorange_threshold)[0][0]
        # take the maximum between this index and 1 to skip the left outlier bin
        idx_v0 = np.max((1, idx_v0))
        # get the left boundary of this bin (hence the [:-1] to get the left boundaries)
        v0 = bins[:-1][idx_v0]
        # get the last index where the histogram backward cumulative sum is greater than threshold
        # (backward hence hist[::-1], and to get the correct index in the backward cumulative sum
        # has to be inversed again hence the second [::-1])
        idx_v1 = np.where(np.cumsum(hist[::-1])[::-1] > self.autorange_threshold)[0][-1]
        # take the minimum between this index and the len -2 to skip the right outlier bin
        idx_v1 = np.min((len(hist) - 2, idx_v1))
        # get the right boundary of this bin (hence the [1:] to get the right boundaries)
        v1 = bins[1:][idx_v1]
        return (v0, v1)

    @Slot()
    def autorange_total(self) -> None:
        hist = self.total_hist_plot.opts.get('y1')
        if len(hist) == 0:
            return
        bins = np.empty(len(hist)+1)
        bins[:-1] = self.total_hist_plot.opts.get('x0')
        bins[-1] = self.total_hist_plot.opts.get('x1')[-1]
        v0, v1 = self.autorange_from_hist(hist, bins)
        self.histogram_widget.setLevels(v0, v1)

    @Slot()
    def autorange_band(self) -> None:
        hist = self.band_hist_plot.opts.get('y1')
        if len(hist) == 0:
            return
        bins = np.empty(len(hist)+1)
        bins[:-1] = self.band_hist_plot.opts.get('x0')
        bins[-1] = self.band_hist_plot.opts.get('x1')[-1]
        v0, v1 = self.autorange_from_hist(hist, bins)
        self.histogram_widget.setLevels(v0, v1)

    # connected to Loader.data_units_loaded
    @Slot(str)
    def set_data_units(self, units: str) -> None:
        self.histogram_widget.axis.setLabel("LOS Displacement", units=units)

    @Slot(float, float)
    def set_v0_v1(self, v0: float, v1: float) -> None:
        self.histogram_widget.setLevels(v0, v1)

    def set_default_colormap(self) -> None:
        # greyscale is the first colormap action (see colormaps.py)
        self.histogram_widget.gradient.menu.actions()[0].trigger()

    @Slot(str)
    def set_colormap(self, name: str) -> None:
        for action in self.histogram_widget.gradient.menu.actions():
            if action.menu():
                for sub_action in action.menu().actions():
                    if isinstance(sub_action.data(), PrivateActionData):
                        if self.histogram_widget.gradient.menu.actionDataToColorMap(sub_action.data()).name == name:
                            sub_action.trigger()
                            return
            elif isinstance(action.data(), PrivateActionData):
                if self.histogram_widget.gradient.menu.actionDataToColorMap(action.data()).name == name:
                    action.trigger()
                    return
        logger.warning(f"colormap {name} not found, switch to greyscale")
        self.set_default_colormap()


class myHistogramWidget(HistogramLUTWidget):

    def __init__(self, parent=None):
        super().__init__(parent=parent, levelMode='mono', gradientPosition='right',
                         orientation='vertical')
        # remove the default gradient menu of pyqtgraph.GradientEditorItem
        del self.gradient.menu
        # create custom gradient menu with chosen default colormaps + our owns
        self.gradient.menu = ColorMapMenu(userList=my_colormaps)
        # remove the None action (that selects a greyscale colormap)
        self.gradient.menu.removeAction(self.gradient.menu.actions()[0])
        # add cyclic colormaps submenu
        cyclic_submenu = self.gradient.menu.addMenu("cyclic")
        for cmap in my_cyclic_colormaps:
            buildMenuEntryAction(cyclic_submenu, cmap.name, cmap)
        # connect gradient menu action as in pyqtgraph.GradientEditorItem
        self.gradient.menu.sigColorMapTriggered.connect(self.gradient.colorMapMenuClicked)
        # remove the ticks of gradient
        self.gradient.showTicks(False)
        # set minimum width so that colorbar is not hidden when minimized
        self.setMinimumWidth(115)
        self.setMaximumWidth(250)
        # move the region indicator to background
        self.region.setZValue(-1)
        # remove the default histogram curve
        self.vb.removeItem(self.plot)

    def changeEvent(self, event: QEvent):
        super().changeEvent(event)
        if event.type() == QEvent.Type.EnabledChange:
            if self.isEnabled():
                self.setForegroundBrush(Qt.GlobalColor.transparent)
            else:
                self.setForegroundBrush(QColor(128, 128, 128, 80))


class myLegendWidget(GraphicsView):

    def __init__(self, parent=None):
        super().__init__(parent=parent)

    def changeEvent(self, event: QEvent):
        super().changeEvent(event)
        if event.type() == QEvent.Type.EnabledChange:
            if self.isEnabled():
                self.setForegroundBrush(Qt.GlobalColor.transparent)
            else:
                self.setForegroundBrush(QColor(128, 128, 128, 80))



# -*- coding: utf-8 -*-

from typing import Any, Optional, Union, TYPE_CHECKING

import logging

import warnings

from collections import OrderedDict

import asyncio
import aiohttp

from PySide6.QtCore import Slot, Signal

from PySide6.QtWidgets import QApplication

from PySide6.QtGui import QPainter, QMatrix4x4

from PySide6.QtOpenGL import (
    QOpenGLShader, QOpenGLTexture, QAbstractOpenGLFunctions, QOpenGLVertexArrayObject
)

from qasync import asyncSlot

from shiboken6 import VoidPtr

from OpenGL import GL

import numpy as np

import rasterio

from insarviz.map.layers.Layer import OpenGLLayer

from insarviz.map.Shaders import VERT_SHADER, ALPHA_SHADER, IMAGE_RGB_SHADER

from insarviz.linalg import matrix

import insarviz.version

if TYPE_CHECKING:
    from insarviz.map.MapModel import MapModel
    from insarviz.map.AbstractMapView import AbstractMapView

logger = logging.getLogger(__name__)


class TileLayer(OpenGLLayer):
    """
    Abstract class
    """

    tiles_cache_size: int = 100
    # user agent required for OpenStreetMap https://operations.osmfoundation.org/policies/tiles/
    http_headers = {
        "User-Agent": f"InsarViz/{insarviz.version.__version__} aiohttp/{aiohttp.__version__}"}

    def __init__(self, name: str, model: "MapModel", map_widget: "AbstractMapView"):
        super().__init__(name)
        self.main_transform = model.loader.dataset.transform
        self.main_crs = model.loader.dataset.crs
        map_widget.displayed_area_changed.connect(self.update_tileset)
        # implements a max_sized FIFO cache, keys are (zoom, x, y)
        self.tiles: OrderedDict[tuple[int, int, int], asyncio.Task] = OrderedDict()
        self.tiles_cache_size = 100
        self.zoom: int = 0
        # dictionnary linking zoom level to pixel ratio from tile to main dataset texture ?
        self.tile_to_texture_pixelratio: dict[int, float] = {}
        self.left_tile: int = 0
        self.right_tile: int = 0
        self.top_tile: int = 0
        self.bottom_tile: int = 0
        self.http_session = aiohttp.ClientSession(headers=TileLayer.http_headers)
        self.crs: rasterio.crs.CRS

    def tile_shape(self, zoom: int) -> tuple[int, int]:
        """
        Return the shape (width, height) in pixels of a tile at the given zoom level
        """
        raise NotImplementedError

    def tilematrix_xy_to_tilematrix_crs(self, zoom: int, x: int, y: int) -> tuple[float, float, float, float]:
        """
        Return the rect of the tile in self.crs coordinates as left, bottom, right, top
        """
        raise NotImplementedError

    def main_crs_to_tilematrix_xy(self, left: float, bottom: float, right: float, top: float,
                                  zoom: int) -> tuple[int, int, int, int]:
        """
        Return the rect of tiles in xy coordinates as left, bottom, right, top
        """
        raise NotImplementedError

    def request_url(self, zoom: int, x: int, y: int) -> str:
        """
        Return the url to request the tile as a string
        """
        raise NotImplementedError

    # connected to MapView.displayed_area_changed
    @asyncSlot(tuple, tuple, tuple)
    async def update_tileset(self, top_left: tuple[float, float], bottom_right: tuple[float, float],
                             shape: tuple[int, int]) -> None:
        """
        _summary_

        Parameters
        ----------
        top_left : tuple[float, float]
            coordinates of the top left corner of MapView in texture coordinates of the data set
        bottom_right :tuple[float, float]
            coordinates of the bottom right corner of MapView in texture coordinates of the data set
        shape : tuple[int, int]
            shape of map view (width, height) in pixels
        """
        screen_to_texture_pixelratio = (np.abs(bottom_right[0] - top_left[0])) / shape[0]
        # set zoom to the nearest zoom in the tilematrix
        self.zoom = min(self.tile_to_texture_pixelratio,
                        key=lambda i: np.abs(screen_to_texture_pixelratio -
                                             self.tile_to_texture_pixelratio.get(i)))
        # transform into x,y tile coordinates
        [left, right], [top, bottom] = rasterio.transform.xy(self.main_transform,
                                                             [top_left[1], bottom_right[1]],
                                                             [top_left[0], bottom_right[0]])
        left, right = min(left, right), max(left, right)
        bottom, top = min(top, bottom), max(top, bottom)
        self.left_tile, self.bottom_tile, self.right_tile, self.top_tile = self.main_crs_to_tilematrix_xy(
            left, bottom, right, top, self.zoom)
        tile_list = []
        for i in range(self.left_tile, self.right_tile+1):
            for j in range(self.top_tile, self.bottom_tile+1):
                if (self.zoom, i, j) in self.tiles:
                    # put the tile at the beginning of the cache if already in here (so it is not
                    # removed when adding the missing tiles)
                    self.tiles.move_to_end((self.zoom, i, j), last=True)
                else:
                    tile_list.append((self.zoom, i, j))
        while len(self.tiles) > self.tiles_cache_size:
            removed = self.tiles.popitem(last=False)
            if removed[1].done() and not removed[1].cancelled():
                if self.context.isValid() and self.offscreen_surface.isValid():
                    self.context.makeCurrent(self.offscreen_surface)
                    # delete the OpenGL textures
                    removed[1].result()[0].destroy()
                    self.context.doneCurrent()
            elif not removed[1].done():
                removed[1].cancel()
        if len(tile_list) > 0:
            for tile in tile_list:
                self.tiles[tile] = asyncio.create_task(self.add_tile(*tile))
            try:
                await asyncio.gather(*[self.tiles[tile] for tile in tile_list])
            except asyncio.CancelledError:
                pass
            self.request_paint.emit()

    async def add_tile(self, zoom: int, x: int, y: int):
        try:
            url = self.request_url(zoom, x, y)
            async with self.http_session.get(url) as resp:
                if resp.status != 200:
                    logger.info(f"HTTP request for tile {url} got {resp.status} status code, abort")
                    return
                tile = await resp.read()
            with warnings.catch_warnings():
                # ignore RuntimeWarning for not georeferenced file
                warnings.filterwarnings("ignore", category=rasterio.errors.NotGeoreferencedWarning,
                                        message=("Dataset has no geotransform, gcps, or rpcs. "
                                                 "The identity matrix will be returned."))
                with rasterio.MemoryFile(tile) as notgeo_tile:
                    with notgeo_tile.open() as src:
                        profile = src.profile
                        if src.count in (3, 4):
                            data = src.read([1, 2, 3])
                        elif src.count == 1:
                            # then maybe the tile is one band with a colormap
                            # idea from https://gis.stackexchange.com/questions/377815/cannot-get-color-plot-of-single-band-tiff-image-with-rasterio
                            try:
                                src.colormap(1)
                            except ValueError:
                                logger.info(f"unsupported format for tile {url}")
                                return
                            # transform the colormap into an array
                            colormap = np.asarray([src.colormap(1).get(i, [0, 0, 0, 0]) for i in range(256)],
                                                  dtype=np.uint8)
                            # reconstruct the data into 3 bands rgb using the colormap
                            data = np.transpose(colormap[src.read(1)][:, :, :3], [2, 0, 1]).copy()
                        else:
                            logger.info(f"unsupported format for tile {url}")
                            return
            tile_width, tile_height = self.tile_shape(zoom)
            left, bottom, right, top = self.tilematrix_xy_to_tilematrix_crs(zoom, x, y)
            transform = rasterio.transform.from_bounds(left, bottom, right, top,
                                                       tile_width, tile_height)
            # TODO which compression to use ?
            profile.update(transform=transform, driver='GTiff', crs=self.crs, compress='lzw',
                           width=tile_width, height=tile_height, count=3)
            if "photometric" in profile:
                del profile["photometric"]
            with rasterio.MemoryFile() as geo_tile:
                with geo_tile.open(**profile) as dataset:
                    dataset.write(data)
                with geo_tile.open() as src:
                    with rasterio.vrt.WarpedVRT(src, crs=self.main_crs) as vrt:
                        img = vrt.read()
                        if img.dtype == "uint8":
                            img = img.astype(np.float32)/255
                        # transpose to change shape from (bands, rows, columns) to (rows, columns, bands)
                        # copy to ensure numpy does not create a view (that opengl cannot read properly)
                        img = np.transpose(img, [1, 2, 0]).copy().astype(np.float32)
                        # model matrix transforms coordinates inside the square (0,1) to coordinates in
                        # pixels relatively to the dataset
                        tile_model_matrix = matrix.from_rasterio_Affine(
                            ~self.main_transform
                            * vrt.transform
                            * rasterio.Affine.scale(*vrt.shape[::-1]))
                        h, w, d = img.shape
                        self.context.makeCurrent(self.offscreen_surface)
                        texture = QOpenGLTexture(QOpenGLTexture.Target.Target2D)
                        texture.setMagnificationFilter(QOpenGLTexture.Filter.Linear)
                        texture.setMinificationFilter(QOpenGLTexture.Filter.LinearMipMapLinear)
                        texture.setWrapMode(QOpenGLTexture.WrapMode.ClampToEdge)
                        texture.setSize(w, h)
                        texture.setFormat(QOpenGLTexture.TextureFormat.RGB32F)
                        texture.allocateStorage(QOpenGLTexture.PixelFormat.RGB,
                                                QOpenGLTexture.PixelType.Float32)
                        texture.setData(QOpenGLTexture.PixelFormat.RGB,
                                        QOpenGLTexture.PixelType.Float32, img.data)
                        texture.generateMipMaps()
                        self.context.doneCurrent()
                        assert texture.textureId(
                        ) != 0, f"TileLayer {self.name} cannot load image texture in OpenGl"
            return (texture, tile_model_matrix)
        except asyncio.CancelledError:
            try:
                texture
                self.context.makeCurrent(self.offscreen_surface)
                texture.destroy()
                self.context.doneCurrent()
            except UnboundLocalError:
                # texture does not exist
                pass
            raise

    def show(self, view_matrix: matrix.Matrix, projection_matrix: matrix.Matrix, show_params: OpenGLLayer.ShowParams,
             painter: Optional[QPainter] = None, vao: Optional[QOpenGLVertexArrayObject] = None,
             glfunc: Optional[QAbstractOpenGLFunctions] = None, blend: bool = True) -> None:
        if painter is not None:
            painter.beginNativePainting()
            # a VAO is required because QPainter bound its own VAO so we need to bind back our own
            assert vao is not None, "OpenGLLayer: vao is required when using QPainter"
        if blend:
            glfunc.glEnable(GL.GL_BLEND)
            glfunc.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA)
        if vao is not None:
            vao.bind()
        self.program.bind()
        for i in range(self.left_tile, self.right_tile+1):
            for j in range(self.top_tile, self.bottom_tile+1):
                if self.tiles.get((self.zoom, i, j), None) is not None:
                    tile = self.tiles.get((self.zoom, i, j))
                    if tile.done() and not tile.cancelled() and not tile.exception():
                        texture, model_matrix = tile.result()
                        # bind textures to texture units
                        glfunc.glActiveTexture(GL.GL_TEXTURE0)
                        texture.bind()
                        # set model, view and projection matrixes
                        self.program.setUniformValue(self.program.uniformLocation('model_matrix'),
                                                     QMatrix4x4(matrix.flatten(model_matrix)))
                        self.program.setUniformValue(self.program.uniformLocation('view_matrix'),
                                                     QMatrix4x4(matrix.flatten(view_matrix)))
                        self.program.setUniformValue(self.program.uniformLocation('projection_matrix'),
                                                     QMatrix4x4(matrix.flatten(projection_matrix)))
                        # set alpha value
                        self.program.setUniformValue1f(self.program.uniformLocation('alpha'),
                                                       self.alpha)
                        # draw the two triangles of the VAO that form a square
                        glfunc.glDrawElements(GL.GL_TRIANGLES, 6, GL.GL_UNSIGNED_INT, VoidPtr(0))
                        texture.release()
                elif self.tiles.get((self.zoom-1, i//2, j//2), None) is not None:
                    pass  # TODO
                else:
                    for k in range(2):
                        for l in range(2):
                            if self.tiles.get((self.zoom+1, 2*i+k, 2*j+l), None) is not None:
                                tile = self.tiles.get((self.zoom+1, 2*i+k, 2*j+l))
                                if tile.done() and not tile.cancelled() and not tile.exception():
                                    texture, model_matrix = tile.result()
                                    # bind textures to texture units
                                    glfunc.glActiveTexture(GL.GL_TEXTURE0)
                                    texture.bind()
                                    # set model, view and projection matrixes
                                    self.program.setUniformValue(self.program.uniformLocation('model_matrix'),
                                                                 QMatrix4x4(matrix.flatten(model_matrix)))
                                    self.program.setUniformValue(self.program.uniformLocation('view_matrix'),
                                                                 QMatrix4x4(matrix.flatten(view_matrix)))
                                    self.program.setUniformValue(self.program.uniformLocation('projection_matrix'),
                                                                 QMatrix4x4(matrix.flatten(projection_matrix)))
                                    # set alpha value
                                    self.program.setUniformValue1f(self.program.uniformLocation('alpha'),
                                                                   self.alpha)
                                    # draw the two triangles of the VAO that form a square
                                    glfunc.glDrawElements(GL.GL_TRIANGLES, 6,
                                                          GL.GL_UNSIGNED_INT, VoidPtr(0))
                                    texture.release()
        self.program.release()
        if vao is not None:
            vao.release()
        if painter is not None:
            painter.endNativePainting()

    def build_program(self) -> None:
        self.program.addShaderFromSourceCode(QOpenGLShader.ShaderTypeBit.Vertex, VERT_SHADER)
        self.program.addShaderFromSourceCode(QOpenGLShader.ShaderTypeBit.Fragment, ALPHA_SHADER)
        self.program.addShaderFromSourceCode(QOpenGLShader.ShaderTypeBit.Fragment, IMAGE_RGB_SHADER)
        self.program.link()
        self.program.bind()
        self.program.setUniformValue1i(self.program.uniformLocation('image'), 0)
        self.program.release()

    def __del__(self):
        """
        Free textures and shader program from the VRAM when the layer is destroyed to prevent
        memory leaks.
        """
        f = asyncio.create_task(self.http_session.close())
        while not f.done():
            QApplication.instance().processEvents()
        try:
            if self.context.isValid() and self.offscreen_surface.isValid():
                self.context.makeCurrent(self.offscreen_surface)
                # delete the OpenGL textures
                for tile in self.tiles.values():
                    if tile.done() and not tile.cancelled():
                        tile.result()[0].destroy()
                # delete the OpenGL shaders program
                del self.program
                self.context.doneCurrent()
        except RuntimeError:
            # the context has already been deleted
            pass



# -*- coding: utf-8 -*-

# imports ##########################################################################################

from typing import Any, Optional

import pathlib

from PySide6.QtCore import Qt, QRectF, QPointF, QEasingCurve

from PySide6.QtGui import QColor, QIcon, QPainter, QPainterPath, QPen

from PySide6.QtOpenGL import QAbstractOpenGLFunctions, QOpenGLVertexArrayObject

from insarviz.map.layers.Layer import Layer

from insarviz.map.TreeItemAttribute import TreeItemColorAttribute, TreeItemIntegerAttribute

from insarviz.map.TreeModel import TreeItem, TreeModel

from insarviz.linalg import matrix, vector

from insarviz import bresenham

from insarviz.Roles import Roles


# SelectionLayer class #############################################################################

class SelectionLayer(Layer):
    removable: bool = False
    renamable: bool = False
    kind: str = "selection layer"

    def __init__(self):
        super().__init__("Selection Layer")
        self.points_folder = SelectionFolder("Points", icon=QIcon('icons:points.png'))
        self.add_child(self.points_folder, 0)
        self.profiles_folder = SelectionFolder("Spatial profiles", icon=QIcon('icons:profile.png'))
        self.add_child(self.profiles_folder, 1)
        self.references_folder = SelectionFolder("References", icon=QIcon('icons:ref.png'))
        self.add_child(self.references_folder, 2)

    def data(self, column: int, role: int) -> Any:
        if column == 0 and role == Qt.ItemDataRole.ToolTipRole:
            return "Selection layer, points, spatial profiles and references"
        return super().data(column, role)

    def show(self, view_matrix: matrix.Matrix, projection_matrix: matrix.Matrix, show_params: Layer.ShowParams,
             painter: Optional[QPainter] = None, vao: Optional[QOpenGLVertexArrayObject] = None,
             glfunc: Optional[QAbstractOpenGLFunctions] = None, blend: bool = True) -> None:
        # pylint: disable=unused-argument
        assert painter is not None
        for reference in self.references_folder.child_items:
            if reference.visible:
                reference.show(view_matrix, painter)
        for profile in self.profiles_folder.child_items:
            if profile.visible:
                profile.show(view_matrix, painter)
        for point in self.points_folder.child_items:
            if point.visible:
                point.show(view_matrix, painter)

    def to_dict(self, project_path: pathlib.Path) -> dict[str, Any]:
        output: dict[str, Any] = super().to_dict(project_path)
        output["points"] = [point.to_dict() for point in self.points_folder.child_items]
        output["profiles"] = [profile.to_dict() for profile in self.profiles_folder.child_items]
        output["references"] = [ref.to_dict() for ref in self.references_folder.child_items]
        return output

    @classmethod
    def from_dict(cls, input_dict: dict[str, Any]) -> "SelectionLayer":
        assert input_dict["kind"] == cls.kind
        selection_layer = SelectionLayer()
        if "visible" in input_dict:
            selection_layer.visible = input_dict["visible"]
        for point in input_dict["points"]:
            selection_layer.points_folder.add_child(SelectionPoint.from_dict(point))
        for profile in input_dict["profiles"]:
            selection_layer.profiles_folder.add_child(SelectionProfile.from_dict(profile))
        for ref in input_dict["references"]:
            selection_layer.references_folder.add_child(SelectionReference.from_dict(ref))
        return selection_layer


# SelectionFolder class ############################################################################

class SelectionFolder(TreeItem):

    def __init__(self, name: str = "", icon: Optional[QIcon] = None):
        super().__init__()
        self.name = name
        if icon is None:
            icon = QIcon()
        self.icon = icon

    def data(self, column: int, role: int):
        if column == TreeModel.remove_column:
            return None
        if role == Qt.ItemDataRole.DisplayRole:
            return self.name
        if role == Qt.ItemDataRole.DecorationRole:
            return self.icon
        if role == Roles.FolderRole:
            # used by LayerView.ProxyLayerModel
            return "SELECTION_FOLDER"
        return None

    def set_data(self, value: Any, column: int, role: int) -> bool:
        return False

    def flags(self, flags: Qt.ItemFlags, column: int) -> Qt.ItemFlags:
        return flags


# SelectionItem classes ############################################################################

def location_path() -> QPainterPath:
    path = QPainterPath()
    path.moveTo(38., 80.)
    path.cubicTo(35., 62., 26., 60., 20., 50.)
    path.cubicTo(0., 0., 80., 0., 60., 50.)
    path.cubicTo(54., 60., 48., 62., 45., 80.)
    path.lineTo(38., 80.)
    return path


def easing_location() -> QEasingCurve:
    easing = QEasingCurve(QEasingCurve.Type.BezierSpline)
    easing.addCubicBezierSegment(QPointF(0.6, 1.), QPointF(0.7, 1.), QPointF(1., 1.))
    return easing


def world_to_view(rect: tuple[int, int, int, int], view_matrix: matrix.Matrix) -> QRectF:
    (left, top, right, bottom) = rect
    left_top_data = vector.matrix((float(left), float(top), 0., 1.))
    left_top_view = matrix.product(view_matrix, left_top_data)
    left_top_view = left_top_view[0][0], left_top_view[1][0]
    # we use right+1 and bottom+1 because we want to fill up to the start of the next pixel
    right_bottom_data = vector.matrix((float(right+1), float(bottom+1), 0., 1.))
    right_bottom_view = matrix.product(view_matrix, right_bottom_data)
    right_bottom_view = right_bottom_view[0][0], right_bottom_view[1][0]
    return QRectF(QPointF(*left_top_view), QPointF(*right_bottom_view))


class SelectionItem(TreeItem):

    ShowCurveRole: int = int(Roles.ShowCurveRole)
    icon: QIcon = QIcon()
    location_icon: QPainterPath = location_path()
    easing_function_loc_icon = easing_location()
    loc_icon_scaling = 0.5
    loc_icon_area_threshold = 25
    loc_icon_area_end_easing = 4

    def __init__(self, name: str = ""):
        super().__init__()
        self.name: str = name
        self.visible: bool = True
        self.show_curve: bool = True
        self.color: QColor = QColor()

    def data(self, column: int, role: int) -> Any:
        if column == TreeModel.remove_column:
            return None
        if role in (Qt.ItemDataRole.DisplayRole, Qt.ItemDataRole.EditRole):
            return self.name
        if role == Qt.ItemDataRole.DecorationRole:
            return self.color
        if role == Qt.ItemDataRole.CheckStateRole:
            return Qt.CheckState.Checked if self.visible else Qt.CheckState.Unchecked
        if role == self.ShowCurveRole:
            return Qt.CheckState.Checked if self.show_curve else Qt.CheckState.Unchecked
        return None

    def flags(self, flags: Qt.ItemFlags, column: int) -> Qt.ItemFlags:
        if column == 0:
            # UserCheckable => visibility, Editable => change name
            flags = flags | Qt.ItemFlag.ItemIsUserCheckable | Qt.ItemFlag.ItemIsEditable
        # Editable => remove
        return flags | Qt.ItemFlag.ItemIsEditable

    def set_data(self, value: Any, column: int, role: int) -> bool:
        if column != 0:
            return False
        if role == Qt.ItemDataRole.CheckStateRole:
            self.visible = (value == Qt.CheckState.Checked or value == Qt.CheckState.Checked.value)
            return True
        if role == self.ShowCurveRole:
            self.show_curve = (value == Qt.CheckState.Checked or value == Qt.CheckState.Checked.value)
            return True
        if role == Qt.ItemDataRole.EditRole:
            self.name = value
            return True
        return False

    def show(self, view_matrix: matrix.Matrix, painter: QPainter, preview: bool = False) -> None:
        """
        Display the layer using either OpenGL commands or painter

        Qpainter documentation on coordinate system : https://doc.qt.io/qt-5/coordsys.html.

        Parameters
        ----------
        view_matrix : matrix.Matrix
            Transform world coordinates into view coordinates.
        painter : QPainter
            QPainter provided by the view.
        preview : bool
            True if the selection item is being constructed
            False if the selection item is in the selection layer
        """
        rect_list = self.get_rect()
        for rect in rect_list:
            view_rect = world_to_view(rect, view_matrix)
            painter.fillRect(view_rect, self.color)
            # location icon only for single rect
            if len(rect_list) == 1:
                area = view_rect.width() * view_rect.height()
                if area < self.loc_icon_area_threshold:
                    painter.save()
                    painter.setRenderHint(QPainter.RenderHint.Antialiasing, True)
                    painter.setBrush(self.color)
                    painter.setPen(QPen(self.color.darker(140), 2))
                    painter.translate(view_rect.center())
                    if area > self.loc_icon_area_end_easing:
                        easing_progress = (self.loc_icon_area_threshold - area) / \
                            (self.loc_icon_area_threshold - self.loc_icon_area_end_easing)
                        scale = self.loc_icon_scaling * \
                            self.easing_function_loc_icon.valueForProgress(easing_progress)
                    else:
                        scale = self.loc_icon_scaling
                    painter.scale(scale, scale)
                    painter.translate(-QPointF(self.location_icon.boundingRect().center().x(),
                                               self.location_icon.boundingRect().bottom()))
                    painter.drawPath(self.location_icon)
                    self.icon.paint(painter, 24, 18, 32, 32,
                                    alignment=Qt.AlignmentFlag.AlignCenter, state=QIcon.State.On)
                    painter.restore()

    def get_rect(self) -> list[tuple[int, int, int, int]]:
        """
        Shall be implemented by subclasses.
        return (left, top, right, bottom)
        """
        raise NotImplementedError

    def to_dict(self) -> dict[str, Any]:
        output: dict[str, Any] = {}
        output["kind"] = self.kind
        output["name"] = self.name
        output["color"] = self.color.name()
        output["visible"] = self.visible
        output["show_curve"] = self.show_curve
        return output

    def init_from_dict(self, input_dict: dict[str, Any]) -> None:
        self.name = input_dict["name"]
        self.color = QColor(input_dict["color"])
        self.visible = input_dict["visible"]
        self.show_curve = input_dict["show_curve"]


class SelectionPoint(SelectionItem):

    kind: str = "point"
    count: int = 0

    def __init__(self, x: int, y: int, r: int):
        SelectionPoint.count += 1
        if self.icon.isNull():
            # cannot be initialized in the class declaration because:
            # "QIcon needs a QGuiApplication instance before the icon is created." see QIcon doc
            SelectionPoint.icon = QIcon('icons:points.png')
        super().__init__(f"Point {self.count}")
        self.x: int = x
        self.y: int = y
        self.r: int = r
        self.add_child(TreeItemIntegerAttribute(self, "r", name="radius",
                                                tooltip="radius in pixels", vmin=1, vmax=1000,
                                                unit="px", datarole=Roles.ComputeDataRole))
        self.color = QColor("blue")
        self.add_child(TreeItemColorAttribute(self, "color", tooltip="color",
                                              datarole=Roles.CurveColorRole))

    def data(self, column: int, role: int) -> Any:
        if role == Qt.ItemDataRole.ToolTipRole:
            return f"Point at ({self.x},{self.y})"
        return super().data(column, role)

    def get_rect(self) -> list[tuple[int, int, int, int]]:
        return [(self.x-self.r+1, self.y-self.r+1, self.x+self.r-1, self.y+self.r-1)]

    def set_r(self, r: int) -> None:
        """Set radius."""
        self.r = r

    def to_dict(self) -> dict[str, Any]:
        output: dict[str, Any] = super().to_dict()
        output["x"] = self.x
        output["y"] = self.y
        output["r"] = self.r
        return output

    @classmethod
    def from_dict(cls, input_dict: dict[str, Any]) -> "SelectionPoint":
        assert input_dict["kind"] == cls.kind
        point = cls(input_dict["x"], input_dict["y"], input_dict["r"])
        point.init_from_dict(input_dict)
        return point


class SelectionProfile(SelectionItem):

    kind: str = "profile"
    count: int = 0
    thickness_threshold = 2
    TemporalRole = Roles.ProfileTemporalRole
    SpatialRole = Roles.ProfileSpatialRole
    temporal_icon: QIcon = QIcon()
    spatial_icon: QIcon = QIcon()

    def __init__(self, x: int, y: int, r: int):
        SelectionProfile.count += 1
        if self.icon.isNull():
            # cannot be initialized in the class declaration because:
            # "QIcon needs a QGuiApplication instance before the icon is created." see QIcon doc
            SelectionProfile.icon = QIcon('icons:profile.png')
            SelectionProfile.temporal_icon = QIcon('icons:temporal.svg')
            SelectionProfile.spatial_icon = QIcon('icons:spatial.svg')
        super().__init__(f"Profile {self.count}")
        self.points: list[tuple[int, int]] = [(x, y)]
        self.temporal_point: Optional[int] = None
        self.spatial_point: Optional[int] = None
        self.r: int = r
        self.add_child(TreeItemIntegerAttribute(self, "r", name="radius",
                                                tooltip="radius in pixels", vmin=1, vmax=1000,
                                                unit="px", datarole=Roles.ComputeDataRole))
        self.color = QColor("orchid")
        self.add_child(TreeItemColorAttribute(self, "color", tooltip="color",
                                              datarole=Roles.CurveColorRole))

    def data(self, column: int, role: int) -> Any:
        if role == Qt.ItemDataRole.ToolTipRole:
            point_list: str = ', '.join('({}, {})'.format(*p) for p in self.points)
            return f"Profile {point_list}"
        return super().data(column, role)

    def set_data(self, value: Any, column: int, role: int) -> bool:
        if role == self.TemporalRole:
            self.temporal_point = value
            return True
        if role == self.SpatialRole:
            self.spatial_point = value
            return True
        return super().set_data(value, column, role)

    def get_rect(self) -> list[tuple[int, int, int, int]]:
        result = []
        for i in range(1, len(self.points)):
            for k in bresenham.line(self.points[i-1][0], self.points[i-1][1], self.points[i][0],
                                    self.points[i][1]):
                result.append((k[0]-self.r+1, k[1]-self.r+1, k[0]+self.r-1, k[1]+self.r-1))
        return result

    def show(self, view_matrix: matrix.Matrix, painter: QPainter, preview: bool = False) -> None:
        for rect in self.get_rect():
            view_rect = world_to_view(rect, view_matrix)
            if view_rect.width() < self.thickness_threshold:
                thickness_change = (self.thickness_threshold - view_rect.width())/2
                view_rect.adjust(-thickness_change, -thickness_change,
                                 thickness_change, thickness_change)
            painter.fillRect(view_rect, self.color)
        if self.temporal_point is not None:
            assert self.temporal_point in range(len(self.get_rect()))
            view_rect = world_to_view(self.get_rect()[self.temporal_point], view_matrix)
            painter.save()
            painter.setRenderHint(QPainter.RenderHint.Antialiasing, True)
            painter.setBrush(self.color)
            painter.setPen(QPen(self.color.darker(140), 2))
            painter.translate(view_rect.center())
            scale = self.loc_icon_scaling
            painter.scale(scale, scale)
            painter.translate(-QPointF(self.location_icon.boundingRect().center().x(),
                                       self.location_icon.boundingRect().bottom()))
            painter.drawPath(self.location_icon)
            self.temporal_icon.paint(painter, 24, 18, 32, 32,
                                     alignment=Qt.AlignmentFlag.AlignCenter, state=QIcon.State.On)
            painter.restore()
        if self.spatial_point is not None:
            assert self.spatial_point in range(len(self.get_rect()))
            view_rect = world_to_view(self.get_rect()[self.spatial_point], view_matrix)
            painter.save()
            painter.setRenderHint(QPainter.RenderHint.Antialiasing, True)
            painter.setBrush(self.color)
            painter.setPen(QPen(self.color.darker(140), 2))
            painter.translate(view_rect.center())
            scale = self.loc_icon_scaling
            painter.scale(scale, scale)
            painter.translate(-QPointF(self.location_icon.boundingRect().center().x(),
                                       self.location_icon.boundingRect().bottom()))
            painter.drawPath(self.location_icon)
            self.spatial_icon.paint(painter, 24, 18, 32, 32,
                                    alignment=Qt.AlignmentFlag.AlignCenter, state=QIcon.State.On)
            painter.restore()

    def add_point(self, x: int, y: int) -> None:
        self.points.append((x, y))

    def remove_last_point(self) -> None:
        del self.points[-1]

    def to_dict(self) -> dict[str, Any]:
        output: dict[str, Any] = super().to_dict()
        output["points"] = self.points
        output["r"] = self.r
        return output

    @classmethod
    def from_dict(cls, input_dict: dict[str, Any]) -> "SelectionProfile":
        assert input_dict["kind"] == cls.kind
        assert len(input_dict["points"]) > 1
        profile = cls(input_dict["points"][0][0], input_dict["points"][0][1], input_dict["r"])
        for point in input_dict["points"][1:]:
            profile.add_point(point[0], point[1])
        profile.init_from_dict(input_dict)
        return profile


class SelectionReference(SelectionItem):

    kind: str = "reference"
    count: int = 0

    def __init__(self, x: int, y: int):
        SelectionReference.count += 1
        if self.icon.isNull():
            # cannot be initialized in the class declaration because:
            # "QIcon needs a QGuiApplication instance before the icon is created." see QIcon doc
            SelectionReference.icon = QIcon('icons:ref.png')
        super().__init__(f"Reference {self.count}")
        self.left: int = x
        self.right: int = x
        self.top: int = y
        self.bottom: int = y
        self.color = QColor("orange")
        self.add_child(TreeItemColorAttribute(self, "color", tooltip="color",
                                              datarole=Roles.CurveColorRole))

    def data(self, column: int, role: int) -> Any:
        if role == Qt.ItemDataRole.ToolTipRole:
            return f"Reference from ({self.left}, {self.top}) to ({self.right}, {self.bottom})"
        return super().data(column, role)

    def get_rect(self) -> list[tuple[int, int, int, int]]:
        return [(self.left, self.top, self.right, self.bottom)]

    def set_rect(self, left: int, top: int, right: int, bottom: int) -> None:
        self.left = left
        self.top = top
        self.right = right
        self.bottom = bottom

    def to_dict(self) -> dict[str, Any]:
        output: dict[str, Any] = super().to_dict()
        output["left"] = self.left
        output["right"] = self.right
        output["top"] = self.top
        output["bottom"] = self.bottom
        return output

    @classmethod
    def from_dict(cls, input_dict: dict[str, Any]) -> "SelectionReference":
        assert input_dict["kind"] == cls.kind
        ref = cls(input_dict["left"], input_dict["top"])
        ref.set_rect(input_dict["left"], input_dict["top"], input_dict["right"],
                     input_dict["bottom"])
        ref.init_from_dict(input_dict)
        return ref
